//============================================================================
// BaseSocket.h - Socket abstract base class
//============================================================================

#ifndef __BASESOCKET_H__
#define __BASESOCKET_H__

//============================================================================
class CBaseSocket
{
public:
	CBaseSocket() {}
	virtual ~CBaseSocket() {}

	// Destroy this socket (Usually delete this)
	virtual void Destroy() = 0;

	// Status of a nonblocking connection
	enum ConnectStatus
	{
		STATUS_NotConnected,		// Socket is not connected
		STATUS_Connecting,			// Nonblocking connect in progress
		STATUS_ConnectError,		// An error occursed during a nonblocking connect
		STATUS_Connected			// Connected
	};

	// General socket status
	enum SocketStatus
	{
		STATUS_Ok,					// Everything is fine
		STATUS_Disconnected,		// Socket has been cleanly disconnected
		STATUS_Error				// Connection was forcefully closed. Check error string
	};

	// Perform a nonblocking connect
	virtual bool StartConnect(const char* szHost, unsigned short wPort) = 0;
	virtual ConnectStatus PollConnectComplete() = 0;

	// Disconnect and cleanup the socket
	virtual void Disconnect() = 0;

	// Reset error string, and nullify connect/socket error stats
	virtual void ClearError() = 0;

	// Check the socket is still connected, and read data if possible
	virtual SocketStatus Poll() = 0;

	// Get one line from the buffer
	virtual const char* GetLine() = 0;

	// Get the whole buffer
	virtual size_t GetBufferSize() const = 0;
	virtual const char* GetBuffer() const = 0;

	// Clear the whole buffer
	virtual void ClearBuffer() = 0;

	// Send a buffer
	virtual bool Send(const void* buff, size_t len) = 0;

	// Accessors
	virtual const char* GetError() const = 0;
	virtual bool IsConnected() const = 0;
	virtual bool DataPending() const = 0;
};

#endif // __BASESOCKET_H__
