//============================================================================
// Config.cpp - Class for reading / writing a config file
//============================================================================

#include "Config.h"
#include <stdio.h>

static bool CfgReadLineFromFile(std::string& str, FILE* pFile)
{
char ch;

	str = "";

	// First, Eat '\r' & '\n' //
	do {
		if(fread(&ch,1,1,pFile) != 1) return false;
	} while((ch == '\r') || (ch == '\n'));
	
	// Read up to next '\r' or '\n' //
	do {
		str += ch;
		if(fread(&ch,1,1,pFile) != 1) return (str.length() != 0);
	} while((ch != '\r') && (ch != '\n'));
	return true;
}

inline std::string& CfgTrimWhitespace(std::string& str)
{
	if(str.empty()) return str;
	while((str[0] == ' ') || (str[0] == '\r') || (str[0] == '\n') || (str[0] == '\t'))
	{
		str.erase(str.begin());
		if(str.empty()) return str;
	}
	while((str[str.length()-1] == ' ') || (str[str.length()-1] == '\r') || 
		(str[str.length()-1] == '\n') || (str[str.length()-1] == '\t'))
	{
		str.erase(str.begin()+str.length()-1);
		if(str.empty()) return str;
	}

	return str;
}

//============================================================================
CConfig::CConfig()
{
}

CConfig::~CConfig()
{
}

void CConfig::Reset()
{
	m_theConfig.clear();
}

bool CConfig::Load(const std::string& strFilename, char chComment)
{
std::string strLine;
std::pair<std::string, std::string> thePair;

	Reset();

	// Open File
	FILE* pFile = fopen(strFilename.c_str(), "r");
	if(!pFile)
		return false;

	// Read lines
	while(CfgReadLineFromFile(strLine, pFile))
	{
		if(!strLine.empty() && strLine[0]==chComment)
			continue;
		strLine = CfgTrimWhitespace(strLine);
		if(strLine.length() == 0)
			continue;

		// Split into key & value
		size_t nPos = strLine.find('=');
		if(nPos == std::string::npos)
		{
			thePair.first = strLine;
			thePair.second = "";
		}
		else
		{
			thePair.first = strLine.substr(0,nPos);
			thePair.second = strLine.substr(nPos+1);
			thePair.first = CfgTrimWhitespace(thePair.first);
			thePair.second = CfgTrimWhitespace(thePair.second);
		}

		m_theConfig[thePair.first] = thePair.second;
	}

	// Cleanup
	fclose(pFile);
	return true;
}

bool CConfig::Save(const std::string& strFilename, const std::string& strHeader) const
{
	FILE* pFile = fopen(strFilename.c_str(), "w");
	if(!pFile)
		return false;

	if(!strHeader.empty())
		fprintf(pFile,"%s\n\n", strHeader.c_str());

	for(StringLookupIter it=m_theConfig.begin(); it!=m_theConfig.end(); ++it)
		fprintf(pFile, "%s = %s\n", it->first.c_str(), it->second.c_str());

	fclose(pFile);
	return true;
}

std::string CConfig::GetValue(const std::string& strKey) const
{
	StringLookupIter it = m_theConfig.find(strKey);
	if(it == m_theConfig.end())
		return "";

	return it->second;
}

void CConfig::SetValue(const std::string& strKey, const std::string& strValue)
{
	m_theConfig[strKey] = strValue;
}
