//============================================================================
// DruinkEdit.cpp - Overridden RichEdit window class
//============================================================================
#include "DruinkEdit.h"
#include "Resource/Resource.h"
#include <RichEdit.h>
#include <CommCtrl.h>
#include <windowsx.h>
#include <algorithm>

#pragma comment(lib, "comctl32.lib")

static const int s_nMaxBufferSize = 256;

static const int KEY_None = 0x00;
static const int KEY_Ctrl = 0x01;
static const int KEY_Shift = 0x02;
static const int KEY_Alt = 0x04;

// Fucking Microsoft headers
#ifdef _WIN64
#	define CAST_TYPE	(LONG_PTR)
#else
#	define CAST_TYPE	(LONG)(LONG_PTR)
#endif

//============================================================================
// Globals and local definitions
bool CDruinkEdit::s_bLoadedDlls = false;

inline bool GetHexColour(const char* szTag, COLORREF& rColour)
{
	rColour = 0;
	for(int i=0, nShift=0; i<3; i++, szTag+=2)
	{
		if((szTag[1] >= '0') && (szTag[1] <= '9')) rColour |= (szTag[1]-'0') << nShift;
		else if((szTag[1] >= 'a') && (szTag[1] <= 'f')) rColour |= (szTag[1]-'a'+10) << nShift;
		else if((szTag[1] >= 'A') && (szTag[1] <= 'F')) rColour |= (szTag[1]-'A'+10) << nShift;
		else return false;
		nShift += 4;
		if((szTag[0] >= '0') && (szTag[0] <= '9')) rColour |= (szTag[0]-'0') << nShift;
		else if((szTag[0] >= 'a') && (szTag[0] <= 'f')) rColour |= (szTag[0]-'a'+10) << nShift;
		else if((szTag[0] >= 'A') && (szTag[0] <= 'F')) rColour |= (szTag[0]-'A'+10) << nShift;
		else return false;
		nShift += 4;
	}

	return true;
}

size_t GetNextTag(const std::string& str)
{
	for(DWORD i=0; i<str.length(); ++i)
	{
		if(str[i] == '\x19') return i;	// TA2 PML tag
		if(str[i] == '\x1A') return i;	// Druink Unicode
		else if(str[i]&0x80) return i;	// TA2 tag
	}

	return std::string::npos;
}

//============================================================================
// Main class
CDruinkEdit::CDruinkEdit() :
	m_hInstance(NULL),
	m_hWndParent(NULL),
	m_hWnd(NULL),
	m_hWndPasteBuffer(NULL),
	m_pfnOldProc(NULL),
	m_bReadOnly(true),
	m_strFont("Courier New"),
	m_nFontSize(10),
	m_bOutputHeld(false),
	m_bMinimized(false),
	m_bFocused(false),
	m_bPassword(false),
	m_strBuffer(""),
	m_crDefaultText(RGB(0xd0, 0xd0, 0xd0)),
	m_crDefaultBack(RGB(0, 0, 0)),
	m_bOverURL(false),
	m_hDefaultCursor(NULL),
	m_hLinkCursor(NULL),
	m_hArrowCursor(NULL),
	m_dwFormatBeforeURL(0),
	m_crColourBeforeURLText(0),
	m_crColourBeforeURLBack(0),
	m_dwURLFormat(CFE_UNDERLINE),
	m_crURLText(RGB(255, 255, 255)),
	m_crURLBack(0),
	m_nRTFFormatID(0),
	m_bMenuVisible(false)
{
	m_szURL[0] = 0;
	m_itBufferPos = m_liBuffer.end();
}

void CDruinkEdit::Release()
{
	m_hInstance = NULL;
	m_hWndParent = NULL;
	if(m_hWnd)
	{
		DestroyWindow(m_hWnd);
		SetWindowLongPtr(m_hWnd, GWLP_USERDATA, 0);
		m_hWnd = NULL;
	}
	if(m_hWndPasteBuffer)
	{
		DestroyWindow(m_hWndPasteBuffer);
		m_hWndPasteBuffer = NULL;
	}
	m_pfnOldProc = NULL;
	m_bReadOnly = true;
	m_strFont = "Courier New";
	m_nFontSize = 10;
	m_bOutputHeld = false;
	m_bMinimized = false;
	m_bFocused = false;
	m_bPassword = false;
	DeleteObject(m_hBackgroundBrush);

	m_strBuffer = "";
	m_crDefaultText = RGB(0xd0, 0xd0, 0xd0);
	m_crDefaultBack = RGB(0, 0, 0);

	m_bOverURL = false;
	m_hDefaultCursor = NULL;
	m_hLinkCursor = NULL;
	m_hArrowCursor = NULL;
	m_szURL[0] = 0;
	m_dwFormatBeforeURL = 0;
	m_crColourBeforeURLText = 0;
	m_crColourBeforeURLBack = 0;
	m_dwURLFormat = CFE_UNDERLINE;
	m_crURLText = RGB(255, 255, 255);
	m_crURLBack = 0;
	m_nRTFFormatID = 0;
	m_bMenuVisible = false;

	m_vPMLTags.clear();

	m_liBuffer.clear();
	m_itBufferPos = m_liBuffer.end();
}

bool CDruinkEdit::Create(HINSTANCE hInstance, HWND hWndParent, DWORD x, DWORD y,
						 DWORD w, DWORD h, bool bReadOnly)
{
	// Copy data
	m_hInstance = hInstance;
	m_hWndParent = hWndParent;
	m_bReadOnly = bReadOnly;

	// Load DLLs and init controls if needed
	if(!s_bLoadedDlls)
	{
		InitCommonControls();
		LoadLibrary("RICHED20.DLL");
		s_bLoadedDlls = true;
	}

	// Init shell execute thread
	if(!m_shellExecuteThread.Init())
		return false;

	// Cache cursor handles
	m_hDefaultCursor = (HCURSOR)LoadImage(NULL, IDC_IBEAM, IMAGE_CURSOR, 0, 0, LR_DEFAULTSIZE | LR_SHARED);
	m_hLinkCursor = (HCURSOR)LoadImage(NULL, IDC_HAND, IMAGE_CURSOR, 0, 0, LR_DEFAULTSIZE | LR_SHARED);
	m_hArrowCursor = (HCURSOR)LoadImage(NULL, IDC_ARROW, IMAGE_CURSOR, 0, 0, LR_DEFAULTSIZE | LR_SHARED);

	// Get RTF clipboard format ID
	m_nRTFFormatID = RegisterClipboardFormat(CF_RTF);

	// Create window
	m_hWnd = CreateWindowEx(WS_EX_TRANSPARENT, RICHEDIT_CLASS, "", GetThisWindowStyle(bReadOnly), x, y, w, h,
		hWndParent, NULL, hInstance, NULL);
	if(!m_hWnd)
		return false;
	SetWindowLongPtr(m_hWnd, GWLP_USERDATA, CAST_TYPE this);

	// Create paste buffer
	m_hWndPasteBuffer = CreateWindowEx(0, RICHEDIT_CLASS, "", GetBaseWindowStyle(), 0, 0, 0, 0, hWndParent,
		NULL, hInstance, NULL);
	if(!m_hWndPasteBuffer)
		return false;

	// Subclass RichEdit
	m_pfnOldProc = (WNDPROC)(LONG_PTR)GetWindowLongPtr(m_hWnd, GWLP_WNDPROC);
	SetWindowLongPtr(m_hWnd, GWLP_WNDPROC, CAST_TYPE CDruinkEdit::StaticWindowProc);

	// Various misc setup
	m_hBackgroundBrush = CreateSolidBrush(m_crDefaultBack);
	SendMessage(m_hWnd, EM_SETLIMITTEXT, INT_MAX, 0);
	SetFont(m_strFont, m_nFontSize);

	// Change text colour
	CHARFORMAT2 cfFormat;
	memset(&cfFormat, 0, sizeof(cfFormat));
	cfFormat.cbSize = sizeof(cfFormat);
	cfFormat.dwMask = CFM_COLOR | CFM_BACKCOLOR | CFM_BOLD | CFM_ITALIC | CFM_UNDERLINE | CFM_STRIKEOUT;
	cfFormat.crTextColor = m_crDefaultText;
	cfFormat.crBackColor = m_crDefaultBack;
	SendMessage(m_hWnd, EM_SETCHARFORMAT, SCF_ALL, (LPARAM)&cfFormat);

	// Done
	return true;
}

//============================================================================

int CDruinkEdit::GetIdealWidth()
{
	// Create DC to query
	HDC hDC = CreateCompatibleDC(NULL);
	if(!hDC)
		return -1;

	// Create font
	HFONT hFont = CreateFont(-MulDiv(m_nFontSize, GetDeviceCaps(hDC, LOGPIXELSY), 72),
		0, 0, 0, FW_BOLD, TRUE, TRUE, TRUE, ANSI_CHARSET, OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, FIXED_PITCH | FF_DONTCARE,
		m_strFont.c_str());
	if(!hFont)
	{
		DeleteDC(hDC);
		return -1;
	}

	// Select the font into the DC and get the string width
	HFONT hFontOld = (HFONT)SelectObject(hDC, hFont);
	SIZE sz;
	const char* szString =
		"____________________________________________________________________________________________________";
	GetTextExtentPoint32(hDC, szString, (int)strlen(szString), &sz);

	// Cleanup font and DC
	SelectObject(hDC, hFontOld);
	DeleteObject(hFont);
	DeleteDC(hDC);

	// Load DLLs and init controls if needed
	if(!s_bLoadedDlls)
	{
		InitCommonControls();
		LoadLibrary("RICHED20.DLL");
		s_bLoadedDlls = true;
	}

	// Create temp window so we can get the format rect size
	HWND hWnd = CreateWindow(RICHEDIT_CLASS, "", GetThisWindowStyle(true) & (~WS_VISIBLE) & (~WS_CHILD),
		0, 0, sz.cx, sz.cy, NULL, NULL, GetModuleHandle(NULL), NULL);
	if(!hWnd)
		return -1;

	// Get formatting rectangle size and determine padding
	RECT rc;
	SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rc);
	int nPadding = sz.cx - (rc.right-rc.left);

	// Cleanup window
	DestroyWindow(hWnd);

	// Add padding
	sz.cx += nPadding;

	// Done
	return sz.cx;
}

//============================================================================

void CDruinkEdit::SetDefaultTextColour(COLORREF crText)
{
	m_crDefaultText = crText;
}

void CDruinkEdit::SetDefaultBackColour(COLORREF crBack)
{
	m_crDefaultBack = crBack;
}

//============================================================================
// Mutators
void CDruinkEdit::SetFont(const std::string& strFont, int nSize)
{
	// Copy data
	m_strFont = strFont;
	m_nFontSize = nSize;

	if(m_hWnd)
	{
		// Setup char format
		CHARFORMAT cfFormat;
		memset(&cfFormat, 0, sizeof(cfFormat));
		cfFormat.cbSize = sizeof(cfFormat);
		cfFormat.dwMask = CFM_CHARSET | CFM_FACE | CFM_SIZE;
		cfFormat.bCharSet = ANSI_CHARSET;
		cfFormat.bPitchAndFamily = FIXED_PITCH | FF_DONTCARE;
		cfFormat.yHeight = (nSize*1440)/72;
		strcpy(cfFormat.szFaceName, strFont.c_str());

		// Set char format and goto end of text
		CHARRANGE cr;
		cr.cpMin = INT_MAX;
		cr.cpMax = INT_MAX;
		SendMessage(m_hWnd, EM_SETCHARFORMAT, SCF_ALL, (LPARAM)&cfFormat);
		SendMessage(m_hWnd, EM_EXSETSEL, 0, (LPARAM)&cr);
	}
}

void CDruinkEdit::SetURLTextColour(COLORREF crText)
{
	m_crURLText = crText;
}

void CDruinkEdit::SetURLBackColour(COLORREF crBack)
{
	m_crURLBack = crBack;
}

void CDruinkEdit::SetURLFormat(BYTE byTA2Bits)
{
	// Convert TA2 bits to RichEdit format
	m_dwURLFormat = 0;
	if(byTA2Bits & TA2BIT_BOLD) m_dwURLFormat |= CFE_BOLD;
	if(byTA2Bits & TA2BIT_ITALIC) m_dwURLFormat |= CFE_ITALIC;
	if(byTA2Bits & TA2BIT_UNDERLINE) m_dwURLFormat |= CFE_UNDERLINE;
	if(byTA2Bits & TA2BIT_STRIKEOUT) m_dwURLFormat |= CFE_STRIKEOUT;
}

void CDruinkEdit::SetURLPrefixes(const std::vector<std::string>& vPrefixes)
{
	// Copy prefixes
	m_vURLPrefixes = vPrefixes;

	// Convert all to lower case
	for(size_t i=0; i<m_vURLPrefixes.size(); ++i)
	{
		std::transform(m_vURLPrefixes[i].begin(), m_vURLPrefixes[i].end(),
			m_vURLPrefixes[i].begin(), tolower);
	}
}

void CDruinkEdit::SetIsPassword(bool isPassword)
{
	if(m_bPassword != isPassword)
	{
		// NB: 0x25cf = black circle Unicode character
		SendMessage(m_hWnd, EM_SETPASSWORDCHAR, isPassword ? 0x25cf : 0, 0);
		m_bPassword = isPassword;
	}
}

void CDruinkEdit::PauseOutput(bool bPause)
{
	m_bOutputHeld = bPause;

	// If we just unpaused, send any pending output
	if(!m_bOutputHeld && !m_strBuffer.empty())
		AddText("");
}

std::string CDruinkEdit::GetText() const
{
	std::string strBuff;
	size_t nLen = SendMessage(m_hWnd, WM_GETTEXTLENGTH, 0, 0);
	strBuff.resize(nLen);
	SendMessage(m_hWnd, WM_GETTEXT, nLen+1, (LPARAM)&strBuff[0]);
	return strBuff;
}

void CDruinkEdit::SetText(const std::string& str)
{
	SetWindowText(m_hWnd, str.c_str());
}

std::string CDruinkEdit::GetNextPMLTag()
{
	if(m_vPMLTags.empty())
		return "";
	std::string strRet = m_vPMLTags[0];
	m_vPMLTags.erase(m_vPMLTags.begin());
	return strRet;
}

void CDruinkEdit::ResetPML()
{
	// Wipe all PML tags from the buffer
	for(size_t i=0; i<m_strBuffer.length(); ++i)
	{
		if(m_strBuffer[i] == '\x19')
		{
			m_strBuffer.erase(m_strBuffer.begin()+i);
			--i;
		}
	}
}

//============================================================================
// Text adding
bool CDruinkEdit::AddText(const std::string& strText, bool bInsert)
{
	size_t nPos;
	m_strBuffer += strText;
	if(m_bOutputHeld || m_bMinimized)
		return false;
	if(m_bReadOnly)
		ProcessURLsInBuffer();

	int nInsertPoint;
	bool bSetSelection = true;
	if(bInsert)
	{
		CHARRANGE selection;
		SendMessage(m_hWnd, EM_EXGETSEL, 0, (LPARAM)&selection);
		nInsertPoint = selection.cpMin;
		bSetSelection = false;
	}
	else
		nInsertPoint = INT_MAX;

	do {
		// Find next tag
		nPos = GetNextTag(m_strBuffer);
		if(nPos == std::string::npos)
			break;

		// Was it an extended client tag?
		size_t nEnd;
		if(m_strBuffer[nPos] == '\x19')
		{
			// Find end of tag
			nEnd = m_strBuffer.find('\x19',nPos+1);
			if(nEnd == std::string::npos)
				break;

			// Add tag
			m_vPMLTags.push_back(m_strBuffer.substr(nPos+1, nEnd-nPos-1));

			// Add any text before tag
			if(nPos != 0)
			{
				AddTextImpl(m_strBuffer.substr(0, nPos), nInsertPoint, bSetSelection);
				if(bInsert)
					nInsertPoint += (int)nPos;
				bSetSelection = true;
			}

			// Trim the end of the tag too
			nEnd++;
			if(m_strBuffer[nEnd] == '\r') nEnd++;
			if(m_strBuffer[nEnd] == '\n') nEnd++;
		}

		// Was it a unicode character?
		else if(m_strBuffer[nPos] == '\x1A')
		{
			// Got end of character?
			nEnd = m_strBuffer.find('\x1A',nPos+1);
			if(nEnd == std::string::npos)
				break;

			// Add any text before character
			if(nPos != 0)
			{
				AddTextImpl(m_strBuffer.substr(0, nPos), nInsertPoint, bSetSelection);
				if(bInsert)
					nInsertPoint += (int)nPos;
				bSetSelection = true;
			}

			// Add character
			size_t charLen = nEnd-nPos-1;
			if(charLen <= 8)
			{
				size_t digit = 0;
				size_t shift = (charLen-1) * 4;
				for(size_t i=0; i<charLen; ++i)
				{
					char ch = (char)tolower(m_strBuffer[nPos+1+i]);
					if(ch >= '0' && ch <= '9')
						digit |= (ch-'0') << shift;
					else if(ch >= 'a' && ch <= 'f')
						digit |= (ch-'a'+10) << shift;
					else
					{
						digit = 0;
						break;
					}
					shift -= 4;
				}
				if(digit != 0)
				{
					std::wstring str;
					str += (wchar_t)digit;
					AddTextImpl(str, nInsertPoint, bSetSelection);
				}
			}
			if(bInsert)
				++nInsertPoint;

			// Trim the end of the tag too
			nEnd++;
		}

		// Was it a TA2 tag?
		else if(m_strBuffer[nPos] & 0x80)
		{
			// Append text before tag
			if(nPos != 0)
			{
				AddTextImpl(m_strBuffer.substr(0, nPos), nInsertPoint, bSetSelection);
				if(bInsert)
					nInsertPoint += (int)nPos;
				bSetSelection = true;
				m_strBuffer.erase(m_strBuffer.begin(), m_strBuffer.begin()+nPos);
				nPos = 0;
			}

			// Setup format
			CHARFORMAT2 cfFormat;
			ZeroMemory(&cfFormat, sizeof(cfFormat));
			cfFormat.cbSize = sizeof(cfFormat);
			if(m_strBuffer[0] & TA2BIT_ENABLE)
				cfFormat.dwEffects = CFE_BOLD | CFE_ITALIC | CFE_UNDERLINE | CFE_STRIKEOUT;

			// See what bits are set
			if(m_strBuffer[0] & TA2BIT_BOLD)
				cfFormat.dwMask |= CFM_BOLD;
			if(m_strBuffer[0] & TA2BIT_ITALIC)
				cfFormat.dwMask |= CFM_ITALIC;
			if(m_strBuffer[0] & TA2BIT_UNDERLINE)
				cfFormat.dwMask |= CFM_UNDERLINE;
			if(m_strBuffer[0] & TA2BIT_TEXTCOLOUR)
			{
				cfFormat.dwMask |= CFM_COLOR;
				if(m_strBuffer[0] & TA2BIT_ENABLE)
				{
					// Check we've got enough buffer space
					if(7 > m_strBuffer.length())
						break;

					// Check its a valid colour
					if(!GetHexColour(m_strBuffer.c_str()+1, cfFormat.crTextColor))
					{
						cfFormat.dwMask &= ~CFM_COLOR;
						m_strBuffer[0] &= ~TA2BIT_TEXTCOLOUR;
					}
				}
				else
					cfFormat.crTextColor = m_crDefaultText;
			}
			if(m_strBuffer[0] & TA2BIT_STRIKEOUT)
				cfFormat.dwMask |= CFM_STRIKEOUT;
			if(m_strBuffer[0] & TA2BIT_URL)
			{
				if(m_strBuffer[0] & TA2BIT_ENABLE)
				{
					// Enabling URL mode. Save current format
					CHARFORMAT2 cfOldFormat;
					ZeroMemory(&cfOldFormat, sizeof(cfOldFormat));
					cfOldFormat.cbSize = sizeof(cfOldFormat);
					CHARRANGE crNew;
					crNew.cpMin = nInsertPoint; crNew.cpMax = nInsertPoint;
					SendMessage(m_hWnd, EM_EXSETSEL, 0, (LPARAM)&crNew);
					SendMessage(m_hWnd, EM_GETCHARFORMAT, SCF_SELECTION, (LPARAM)&cfOldFormat);
					m_dwFormatBeforeURL = cfOldFormat.dwEffects;
					m_crColourBeforeURLText = cfOldFormat.crTextColor;
					m_crColourBeforeURLBack = cfOldFormat.crBackColor;

					cfFormat.dwMask = CFM_BOLD | CFM_ITALIC | CFM_UNDERLINE | CFM_STRIKEOUT | CFM_COLOR | CFM_BACKCOLOR;
					cfFormat.dwEffects = m_dwURLFormat;
					if(m_crURLText == 0xffffffff)
						cfFormat.dwMask &= ~CFM_COLOR;
					else
						cfFormat.crTextColor = m_crURLText;
					if(m_crURLBack == 0xffffffff)
						cfFormat.dwMask &= ~CFM_BACKCOLOR;
					else
						cfFormat.crBackColor = m_crURLBack;
				}
				else
				{
					// Disabling URL mode. Restore previous format
					cfFormat.dwMask = CFM_BOLD | CFM_ITALIC | CFM_UNDERLINE | CFM_STRIKEOUT | CFM_COLOR | CFM_BACKCOLOR;
					cfFormat.dwEffects = m_dwFormatBeforeURL;
					cfFormat.crTextColor = m_crColourBeforeURLText;
					cfFormat.crBackColor = m_crColourBeforeURLBack;
				}
			}

			// Set new formatting
			CHARRANGE crNew;
			crNew.cpMin = nInsertPoint; crNew.cpMax = nInsertPoint;
			SendMessage(m_hWnd, EM_EXSETSEL, 0, (LPARAM)&crNew);
			SendMessage(m_hWnd, EM_SETCHARFORMAT, SCF_SELECTION, (LPARAM)&cfFormat);

			// Trim buffer
			if((m_strBuffer[0] & (TA2BIT_ENABLE | TA2BIT_TEXTCOLOUR)) == (TA2BIT_ENABLE | TA2BIT_TEXTCOLOUR))
				nEnd = 7;
			else
				nEnd = 1;
		}
		else
			nEnd = 1;

		// Remove all text prior to tag
		m_strBuffer.erase(m_strBuffer.begin(), m_strBuffer.begin()+nEnd);

	} while(nPos != std::string::npos);

	// If we didn't find a tag, add remaining text
	if((nPos == std::string::npos) && (m_strBuffer.length() > 0))
	{
		AddTextImpl(m_strBuffer, nInsertPoint, bSetSelection);
		m_strBuffer = "";
	}

	return true;
}

void CDruinkEdit::Clear()
{
	SetWindowText(m_hWnd, "");
	m_bOverURL = false;
	m_szURL[0] = 0;
}

//============================================================================
// Message Procs
LRESULT CALLBACK CDruinkEdit::StaticWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
CDruinkEdit* pParent;

	pParent = (CDruinkEdit*)(LONG_PTR)GetWindowLongPtr(hWnd, GWLP_USERDATA);
	if(!pParent) return DefWindowProc(hWnd, uMsg, wParam, lParam);
	return pParent->WindowProc(uMsg, wParam, lParam);
}

LRESULT CDruinkEdit::WindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
static char s_chLast = 0;

	switch(uMsg)
	{
		case WM_DESTROY:
			OnDestroy();
			break;

		case WM_SETFOCUS:
			m_bFocused = true;
			break;

		case WM_KILLFOCUS:
			m_bFocused = false;
			break;

		case WM_ERASEBKGND:
			OnEraseBackground((HDC)wParam);
			return 0;

		case WM_KEYDOWN:
			s_chLast = (char)wParam;
			if(!OnKeyDown((int)wParam, lParam))
				return 0;
			break;

		case WM_KEYUP:
			s_chLast = 0;
			break;

		case WM_CHAR:
		{	
			// Notify parent window on keypress in read only control
			bool bCtrlHeld = !!(GetKeyState(VK_CONTROL)&0x8000);
			if(m_bReadOnly && !bCtrlHeld)
			{
				// Notify parent window
				DruinkEditNotify notify;
				notify.hdr.hwndFrom = m_hWnd;
				notify.hdr.idFrom = 0;
				notify.hdr.code = DEN_ISREADONLY;
				notify.nParam = wParam;
				notify.nParam2 = lParam;
				SendMessage(m_hWndParent, WM_NOTIFY, 0, (LPARAM)&notify);
			}
		}
		break;

		case WM_MOUSEMOVE:
			if(m_bReadOnly)
			{
				// Check for URLs
				m_bOverURL = OverURL(LOWORD(lParam), HIWORD(lParam));
			}
			UpdateCursor();
			break;

		case WM_LBUTTONUP:
			if(m_bOverURL)
				DoShellExecute(m_szURL);
			break;

		case WM_CONTEXTMENU:
			ShowPopupMenu(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			break;

		case WM_INITMENUPOPUP:
			InitPopupMenu((HMENU)wParam);
			m_bMenuVisible = true;
			UpdateCursor();
			return 0;
			break;

		case WM_UNINITMENUPOPUP:
			m_bMenuVisible = false;
			UpdateCursor();
			return 0;
			break;

		case WM_COMMAND:
			if(HIWORD(wParam) == 0)
			{
				OnCommand(LOWORD(wParam));
				return 0;
			}
			break;

		case WM_SETCURSOR:
			UpdateCursor();
			return TRUE;
			break;

		case WM_WINDOWPOSCHANGED:
			OnWindowPosChanged((WINDOWPOS*)lParam);
			break;
	}

	return CallWindowProc(m_pfnOldProc, m_hWnd, uMsg, wParam, lParam);
}

//============================================================================
// Message handlers
void CDruinkEdit::OnDestroy()
{
	SetWindowLongPtr(m_hWnd, GWLP_WNDPROC, CAST_TYPE m_pfnOldProc);
	SetWindowLongPtr(m_hWnd, GWLP_USERDATA, 0);
}

void CDruinkEdit::OnEraseBackground(HDC hDC)
{
	RECT rcClient;
	GetClientRect(m_hWnd, &rcClient);
	FillRect(hDC, &rcClient, m_hBackgroundBrush);
}

bool CDruinkEdit::OnKeyDown(int nVirtKey, LPARAM lKeyData)
{
	// We don't care about any input on a read-only control
	if(m_bReadOnly)
		return true;

	int nKeys = KEY_None;
	if(GetKeyState(VK_CONTROL) & 0x8000) nKeys |= KEY_Ctrl;
	if(GetKeyState(VK_SHIFT) & 0x8000) nKeys |= KEY_Shift;
	if(GetKeyState(VK_MENU) & 0x8000) nKeys |= KEY_Alt;

	if(nVirtKey == '\r' || nVirtKey == '\n')
	{
		// Add string to buffer
		std::string strLine(GetText());
		if(strLine.length() != 0 && !m_bPassword)
		{
			bool bAtEnd = m_itBufferPos == m_liBuffer.end();
			m_liBuffer.push_back(strLine);
			if(m_liBuffer.size() > s_nMaxBufferSize)
				m_liBuffer.erase(m_liBuffer.begin());
			if(bAtEnd)
				m_itBufferPos = m_liBuffer.end();
			else
				++m_itBufferPos;
		}

		// If enter was pressed, notify parent window through WM_NOTIFY
		if((nKeys & (KEY_Ctrl|KEY_Shift)) == 0)
		{
			DruinkEditNotify notify;
			notify.hdr.hwndFrom = m_hWnd;
			notify.hdr.idFrom = 0;
			notify.hdr.code = DEN_NEWLINE;
			notify.nParam = 0;
			SendMessage(m_hWndParent, WM_NOTIFY, 0, (LPARAM)&notify);
			return false;
		}
		return true;
	}
	if(nKeys == KEY_None)
		m_itBufferPos = m_liBuffer.end();
	else if(nKeys == KEY_Ctrl)
	{
		// Handle Ctrl+Up (previous line)
		if((nVirtKey == VK_UP) && (m_liBuffer.size() > 0))
		{
			// Add this line to the buffer if its not already there
			if(m_itBufferPos == m_liBuffer.end())
			{
				std::string strLastLine = *(--m_liBuffer.end());
				std::string strCurrLine(GetText());
				if(((strLastLine.length() > strCurrLine.length()) ||
					(strncmp(strLastLine.c_str(), strCurrLine.c_str(), strLastLine.length()) != 0))
					&& (strCurrLine.length() > 0))
				{
					m_liBuffer.push_back(strCurrLine);
					if(m_liBuffer.size() > s_nMaxBufferSize)
						m_liBuffer.erase(m_liBuffer.begin());
					--m_itBufferPos;
				}
			}

			// Set new text
			if(m_itBufferPos != m_liBuffer.begin())
				--m_itBufferPos;

			SetText(*m_itBufferPos);
			return false;
		}

		// Handle Ctrl+Down (next line)
		else if((nVirtKey == VK_DOWN) && (m_liBuffer.size() > 0))
		{
			// Set new text //
			if(m_itBufferPos != m_liBuffer.end())
			{
				++m_itBufferPos;
				if(m_itBufferPos != m_liBuffer.end())
				{
					SetText(*m_itBufferPos);
				}
				else
				{
					// Add this line to the buffer if its not already there
					std::string strLastLine = *(--m_liBuffer.end());
					std::string strCurrLine(GetText());
					if(((strLastLine.length() > strCurrLine.length()) ||
						(strncmp(strLastLine.c_str(), strCurrLine.c_str(), strLastLine.length()) != 0))
						&& (strCurrLine.length() > 0))
					{
						m_liBuffer.push_back(strCurrLine);
						if(m_liBuffer.size() > s_nMaxBufferSize)
							m_liBuffer.erase(m_liBuffer.begin());
					}

					SetText("");
				}
			}
			else
			{
				// Add this line to the buffer if its not already there
				std::string strLastLine = *(--m_liBuffer.end());
				std::string strCurrLine(GetText());
				if(((strLastLine.length() > strCurrLine.length()) ||
					(strncmp(strLastLine.c_str(), strCurrLine.c_str(), strLastLine.length()) != 0))
					&& (strCurrLine.length() > 0))
				{
					m_liBuffer.push_back(strCurrLine);
					if(m_liBuffer.size() > s_nMaxBufferSize)
						m_liBuffer.erase(m_liBuffer.begin());
				}

				SetText("");
			}

			return false;
		}
	}

	UNREFERENCED_PARAMETER(lKeyData);
	return true;
}

static bool IsURLCharacter(char ch)
{
	return ch > ' ';
}

bool CDruinkEdit::OverURL(WORD x, WORD y)
{
	// Assume not over a URL
	m_szURL[0] = 0;

	// Get character the mouse is over
	POINTL pt = {x, y};
	LRESULT nChar = SendMessage(m_hWnd, EM_CHARFROMPOS, 0, (LPARAM)&pt);

	// Grab range of text around that point
	TEXTRANGE text;
	char szBuff[ms_nURLBufferSize];
	long lMin = (long)(nChar - ms_nURLBufferSize/2 + 1);
	if(lMin < 0) lMin = 0;
	long lMax = (long)(nChar + ms_nURLBufferSize/2 - 1);
	text.chrg.cpMin = lMin;
	text.chrg.cpMax = lMax;
	text.lpstrText = szBuff;
	LRESULT nLen = SendMessage(m_hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&text);
	szBuff[nLen] = 0;
	long nCharPos = (long)(lMin==0 ? nChar : ms_nURLBufferSize/2 - 1);

	// Search back to a space
	long nStart = nCharPos;
	for(; nStart>=0; --nStart)
	{
		if(!IsURLCharacter(szBuff[nStart]))
			break;
	}

	// Search forward to a space
	long nEnd = nCharPos;
	for(; nEnd<nLen; ++nEnd)
	{
		if(!IsURLCharacter(szBuff[nEnd]))
			break;
	}

	// Bail on empty string or too short a string
	if(nEnd - nStart < 4)
		return false;

	// Grab word and convert to lower case
	szBuff[nEnd] = 0;
	const char* pszStart = szBuff + nStart + 1;
	char szLowerBuff[ms_nURLBufferSize];
	for(size_t i=0; pszStart[i]; ++i)
		szLowerBuff[i] = (char)tolower(pszStart[i]);
	szLowerBuff[nEnd-nStart-1] = 0;

	// Check known URL types
	bool bIsURL = false;
	for(size_t i=0; i<m_vURLPrefixes.size(); ++i)
	{
		if(strncmp(szLowerBuff, m_vURLPrefixes[i].c_str(), m_vURLPrefixes[i].length()) == 0)
		{
			bIsURL = true;
			break;
		}
	}
	if(bIsURL)
	{
		// Copy URL to buffer
		strcpy(m_szURL, pszStart);
	}
	return bIsURL;
}

//============================================================================
// Internal functions
void CDruinkEdit::AddTextImpl(const std::string& strText, int nInsertPoint, bool bSetSelection)
{
	if(strText.length() == 0)
		return;

	if(bSetSelection)
	{
		CHARRANGE crNew;
		crNew.cpMin = nInsertPoint; crNew.cpMax = nInsertPoint;
		SendMessage(m_hWnd, EM_EXSETSEL, 0, (LPARAM)&crNew);
	}

	// Insert text
	SETTEXTEX st;
	st.flags = ST_SELECTION;
	if(!m_bReadOnly)
		st.flags |= ST_KEEPUNDO;
	st.codepage = CP_ACP;
	SendMessage(m_hWnd, EM_SETTEXTEX, (WPARAM)&st, (LPARAM)strText.c_str());
}

void CDruinkEdit::AddTextImpl(const std::wstring& strText, int nInsertPoint, bool bSetSelection)
{
	if(strText.length() == 0)
		return;

	if(bSetSelection)
	{
		CHARRANGE crNew;
		crNew.cpMin = nInsertPoint; crNew.cpMax = nInsertPoint;
		SendMessage(m_hWnd, EM_EXSETSEL, 0, (LPARAM)&crNew);
	}

	// Insert text as unicode
	SETTEXTEX st;
	st.flags = ST_SELECTION | 8; //ST_UNICODE;
	if(!m_bReadOnly)
		st.flags |= ST_KEEPUNDO;
	st.codepage = 1200; // Unicode codepage
	SendMessage(m_hWnd, EM_SETTEXTEX, (WPARAM)&st, (LPARAM)strText.c_str());
}

//============================================================================

void CDruinkEdit::ProcessURLsInBuffer()
{
	std::string strLowerBuffer(m_strBuffer);
	std::transform(strLowerBuffer.begin(), strLowerBuffer.end(), strLowerBuffer.begin(), tolower);

	for(size_t i=0; i<m_strBuffer.length(); ++i)
	{
		for(size_t j=0; j<m_vURLPrefixes.size(); ++j)
		{
			if(strncmp(m_vURLPrefixes[j].c_str(), m_strBuffer.c_str()+i, m_vURLPrefixes[j].length()) != 0)
				continue;
			size_t nPos = i;

			// Check it's got whitespace before it
			if(nPos != 0 && IsURLCharacter(strLowerBuffer[nPos-1]))
			{
				// See if it's a TA2 tag
				if((strLowerBuffer[nPos-1]&0x80)==0)
				{
					// See if it's a colour tag (0xc4)
					if(nPos<7 || (strLowerBuffer[nPos-7]&0xc4) != 0xc4)
						continue;
				}
			}

			// Check there's whitespace at the end of the buffer
			size_t nEnd;
			for(nEnd=nPos+1; nEnd<strLowerBuffer.length(); ++nEnd)
			{
				if(!IsURLCharacter(strLowerBuffer[nEnd]) || strLowerBuffer[nEnd] & 0x80)
					break;
			}
			if(nEnd == strLowerBuffer.length())
				continue;

			// Ok, there's a URL here. Insert formatting
			unsigned char chOn = TA2BIT | TA2BIT_ENABLE | TA2BIT_URL;
			unsigned char chOff = TA2BIT | TA2BIT_URL;
			m_strBuffer.insert(m_strBuffer.begin()+nEnd, (char)chOff);
			m_strBuffer.insert(m_strBuffer.begin()+nPos, (char)chOn);
			strLowerBuffer.insert(strLowerBuffer.begin()+nEnd, (char)chOff);
			strLowerBuffer.insert(strLowerBuffer.begin()+nPos, (char)chOn);

			// Skip to end of URL
			i += (nEnd-nPos) + 1;
			break;
		}
	}
}

//============================================================================

void CDruinkEdit::ShowPopupMenu(int x, int y)
{
	// Load menu
	HMENU hMenu = LoadMenu(m_hInstance, MAKEINTRESOURCE(m_bReadOnly?IDM_DRUINKOUTPUT:IDM_DRUINKINPUT));
	if(!hMenu)
		return;
	HMENU hSubMenu = GetSubMenu(hMenu, 0);
	if(!hSubMenu)
	{
		DestroyMenu(hMenu);
		return;
	}

	// Display and track the popup menu
	TrackPopupMenu(hSubMenu, 0, x, y, 0, m_hWnd, NULL);

	DestroyMenu(hMenu);
	DestroyMenu(hSubMenu);
}

void CDruinkEdit::InitPopupMenu(HMENU hMenu)
{
	// See if there's a region selected
	CHARRANGE selection;
	SendMessage(m_hWnd, EM_EXGETSEL, 0, (LPARAM)&selection);

	int nItems = GetMenuItemCount(hMenu);
	for(int i=0; i<nItems; ++i)
	{
		UINT nID = GetMenuItemID(hMenu, i);
		UINT nFlags;
		switch(nID)
		{
			case IDM_CUT:
			case IDM_COPY:
			case IDM_GOOGLE:
			case IDM_COPYTAGGED:
				if(selection.cpMin == selection.cpMax)
					nFlags = MF_BYCOMMAND | MF_GRAYED;
				else
					nFlags = MF_BYCOMMAND | MF_ENABLED;
				EnableMenuItem(hMenu, nID, nFlags);
				break;

			case IDM_COPYURL:
				if(m_bOverURL)
					nFlags = MF_BYCOMMAND | MF_ENABLED;
				else
					nFlags = MF_BYCOMMAND | MF_GRAYED;
				EnableMenuItem(hMenu, nID, nFlags);
				break;

			case IDM_SAVEURL:
				if(m_bOverURL && IsHTTPUrl(m_szURL))
					nFlags = MF_BYCOMMAND | MF_ENABLED;
				else
					nFlags = MF_BYCOMMAND | MF_GRAYED;
				EnableMenuItem(hMenu, nID, nFlags);
				break;

			case IDM_PASTE:
				if(IsClipboardFormatAvailable(m_nRTFFormatID) || IsClipboardFormatAvailable(CF_TEXT))
					nFlags = MF_BYCOMMAND | MF_ENABLED;
				else
					nFlags = MF_BYCOMMAND | MF_GRAYED;
				EnableMenuItem(hMenu, nID, nFlags);
				break;

			case IDM_PASTEPLAIN:
				if(IsClipboardFormatAvailable(CF_TEXT))
					nFlags = MF_BYCOMMAND | MF_ENABLED;
				else
					nFlags = MF_BYCOMMAND | MF_GRAYED;
				EnableMenuItem(hMenu, nID, nFlags);
				break;

			case IDM_HOLD:
			{
				MENUITEMINFO info;
				memset(&info, 0, sizeof(info));
				info.cbSize = sizeof(info);
				info.fMask = MIIM_STATE;
				info.fState = m_bOutputHeld ? MFS_CHECKED : MFS_UNCHECKED;
				SetMenuItemInfo(hMenu, nID, FALSE, &info);
			}
			break;
		}
	}
}

void CDruinkEdit::OnCommand(WORD nID)
{
	switch(nID)
	{
		case IDM_CUT:
			if(!Copy())
				MessageBeep(MB_ICONEXCLAMATION);
			else if(!m_bReadOnly)
				SendMessage(m_hWnd, EM_REPLACESEL, TRUE, (LPARAM)"");
			break;

		case IDM_COPY:
			if(!Copy())
				MessageBeep(MB_ICONEXCLAMATION);
			break;

		case IDM_COPYTAGGED:
			if(!Copy(true))
				MessageBeep(MB_ICONEXCLAMATION);
			break;

		case IDM_PASTE:
			if(!m_bReadOnly)
			{
				if(!Paste())
					MessageBeep(MB_ICONEXCLAMATION);
			}
			break;

		case IDM_PASTEPLAIN:
			if(!m_bReadOnly)
			{
				if(!Paste(true))
					MessageBeep(MB_ICONEXCLAMATION);
			}
			break;

		case IDM_COPYURL:
			if(m_szURL[0]=='\0' || !CopyURL())
				MessageBeep(MB_ICONEXCLAMATION);
			break;

		case IDM_SAVEURL:
			if(m_szURL[0] != '\0')
			{
				// Notify parent window
				DruinkEditNotify notify;
				notify.hdr.hwndFrom = m_hWnd;
				notify.hdr.idFrom = 0;
				notify.hdr.code = DEN_SAVEURL;
				notify.nParam = (UINT_PTR)m_szURL;
				SendMessage(m_hWndParent, WM_NOTIFY, 0, (LPARAM)&notify);
			}
			break;

		case IDM_SELECTALL:
		{
			CHARRANGE cr;
			cr.cpMin = 0;
			cr.cpMax = INT_MAX;
			SendMessage(m_hWnd, EM_EXSETSEL, 0, (LPARAM)&cr);
		}
		break;

		case IDM_HOLD:
		{
			if(m_bReadOnly)
				PauseOutput(!m_bOutputHeld);
		}
		break;

		case IDM_CLEAR:
			SetWindowText(m_hWnd, "");
			break;

		case IDM_BOLD: if(!m_bReadOnly) OnFormat('B'); break;
		case IDM_ITALIC: if(!m_bReadOnly) OnFormat('I'); break;
		case IDM_UNDERLINE: if(!m_bReadOnly) OnFormat('U'); break;
		case IDM_STRIKEOUT: if(!m_bReadOnly) OnFormat('S'); break;

		case IDM_GOOGLE:
			GoogleSearch();
		break;
	}
}

//============================================================================

SIZE CDruinkEdit::GetSizeInCharacters() const
{
	// Get client size
	RECT rcClient;
	GetClientRect(m_hWnd, &rcClient);
	int w = rcClient.right - rcClient.left;
	int h = rcClient.bottom - rcClient.top;

	// Need to create a DC and select the font into it to measure its size,
	// since GetDC(m_hWnd) doesn't give the correct size for some reason
	SIZE charSize = {0, 0};
	HDC hDC = CreateCompatibleDC(NULL);
	if(hDC)
	{
		HFONT hFont = CreateFont(-MulDiv(m_nFontSize, GetDeviceCaps(hDC, LOGPIXELSY), 72),
			0, 0, 0, FW_BOLD, TRUE, TRUE, TRUE, ANSI_CHARSET, OUT_DEFAULT_PRECIS,
			CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, FIXED_PITCH | FF_DONTCARE,
			m_strFont.c_str());
		if(hFont)
		{
			HFONT hFontOld = (HFONT)SelectObject(hDC, hFont);
			GetTextExtentPoint32(hDC, "X", 1, &charSize);
			SelectObject(hDC, hFontOld);
			DeleteObject(hFont);
		}
		DeleteDC(hDC);
	}

	// Fall back to inaccurate method
	if(charSize.cx == 0)
		GetTextExtentPoint32(GetDC(m_hWnd), "X", 1, &charSize);

	// Calculate client size in characters. Minus 2 because of the 1px left and right borders.
	charSize.cx = (w-2) / charSize.cx;
	charSize.cy = (h-2) / charSize.cy;
	return charSize;
}

//============================================================================

static DWORD CALLBACK StreamOutCallback(DWORD_PTR dwCookie, LPBYTE pBuff, LONG cb, LONG* pcb)
{
	std::string& strBuff = *(std::string*)dwCookie;
	size_t nOldLen = strBuff.length();
	strBuff.resize(strBuff.length()+cb);
	memcpy(&strBuff[nOldLen], pBuff, cb);
	*pcb = cb;
	return 0;
}

static bool CopyTextToClipboard(UINT nFormat, const std::string& strBuff)
{
	HGLOBAL hMem = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, strBuff.length() + 1);
	if(!hMem)
		return false;

	void* pData = GlobalLock(hMem);
	if(!pData)
	{
		GlobalFree(hMem);
		return false;
	}
	memcpy(pData, strBuff.c_str(), strBuff.length() + 1);
	GlobalUnlock(hMem);
	SetClipboardData(nFormat, hMem);
	return true;
}

bool CDruinkEdit::Copy(bool bCopyTaggedPlaintext)
{
	// No selection means copy fails
	CHARRANGE selection;
	SendMessage(m_hWnd, EM_EXGETSEL, 0, (LPARAM)&selection);
	if(selection.cpMin == selection.cpMax)
		return false;

	// Streamout to a std::string buffer in RTF format
	EDITSTREAM es;
	std::string strBuff;
	es.dwCookie = (DWORD_PTR)&strBuff;
	es.dwError = 0;
	es.pfnCallback = StreamOutCallback;
	SendMessage(m_hWnd, EM_STREAMOUT, SFF_SELECTION|SF_RTF, (LPARAM)&es);

	// Get selected text as plaintext
	std::string strPlain;
	if(bCopyTaggedPlaintext)
	{
		strPlain = ConvertRTFString(strBuff);
	}
	else
	{
		TEXTRANGE tr;
		SendMessage(m_hWnd, EM_EXGETSEL, 0, (LPARAM)&tr.chrg);
		strPlain.resize(tr.chrg.cpMax - tr.chrg.cpMin);
		tr.lpstrText = &strPlain[0];
		SendMessage(m_hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&tr);
		if(strPlain[strPlain.length()-1] == '\0')
			strPlain.erase(strPlain.begin()+strPlain.length()-1);
	}

	// Open Clipboard
	if(!OpenClipboard(m_hWnd))
		return false;
	EmptyClipboard();

	// Write RTF data
	if(!bCopyTaggedPlaintext)
	{
		if(!CopyTextToClipboard(m_nRTFFormatID, strBuff))
		{
			CloseClipboard();
			return false;
		}
	}

	// Write plain text data
	if(!CopyTextToClipboard(CF_TEXT, strPlain))
	{
		CloseClipboard();
		return false;
	}

	// Cleanup
	CloseClipboard();
	return true;
}

bool CDruinkEdit::CopyURL()
{
	// Open Clipboard
	if(!OpenClipboard(m_hWnd))
		return false;
	EmptyClipboard();

	// Write plain text data
	if(!CopyTextToClipboard(CF_TEXT, m_szURL))
	{
		CloseClipboard();
		return false;
	}

	// Cleanup
	CloseClipboard();
	return true;
}

bool CDruinkEdit::Paste(bool bPlaintextOnly)
{
	// Open Clipboard
	if(!OpenClipboard(m_hWnd))
		return false;

	// Get content
	if(!bPlaintextOnly && IsClipboardFormatAvailable(m_nRTFFormatID))
	{
		// RTF paste
		HANDLE hMemory = GetClipboardData(m_nRTFFormatID);
		if(!hMemory)
		{
			CloseClipboard();
			return false;
		}
		const char* pData = (const char*)GlobalLock(hMemory);
		if(!pData)
		{
			CloseClipboard();
			return false;
		}

		// Convert to tagged string
		AddText(ConvertRTFString(pData), true);

		// Cleanup
		GlobalUnlock(hMemory);
	}
	else
	{
		// Plain text paste
		HANDLE hMemory = GetClipboardData(CF_TEXT);
		if(!hMemory)
		{
			CloseClipboard();
			return false;
		}
		const char* pData = (const char*)GlobalLock(hMemory);
		if(!pData)
		{
			CloseClipboard();
			return false;
		}

		// Paste text
		SendMessage(m_hWnd, EM_REPLACESEL, TRUE, (LPARAM)pData);

		// Cleanup
		GlobalUnlock(hMemory);
	}

	CloseClipboard();
	return true;
}

//============================================================================

bool CDruinkEdit::GoogleSearch()
{
	// Get range selected
	TEXTRANGE tr;
	std::string strBuff;
	SendMessage(m_hWnd, EM_EXGETSEL, 0, (LPARAM)&tr.chrg);
	if(tr.chrg.cpMin == tr.chrg.cpMax)
		return false;

	// Get selected text
	strBuff.resize(tr.chrg.cpMax - tr.chrg.cpMin);
	tr.lpstrText = &strBuff[0];
	SendMessage(m_hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&tr);
	if(strBuff[strBuff.length()-1] == '\0')
		strBuff.erase(strBuff.begin()+strBuff.length()-1);

	// Notify parent window
	DruinkEditNotify notify;
	notify.hdr.hwndFrom = m_hWnd;
	notify.hdr.idFrom = 0;
	notify.hdr.code = DEN_GOOGLE;
	notify.nParam = (UINT_PTR)strBuff.c_str();
	SendMessage(m_hWndParent, WM_NOTIFY, 0, (LPARAM)&notify);
	return true;
}

//============================================================================

void CDruinkEdit::OnFormat(char chFormat)
{
	// Get selected text
	std::string strBuff;
	TEXTRANGE tr;
	SendMessage(m_hWnd, EM_EXGETSEL, 0, (LPARAM)&tr.chrg);
	bool bEmpty = tr.chrg.cpMax == tr.chrg.cpMin;
	strBuff.resize((tr.chrg.cpMax - tr.chrg.cpMin)+4, '*');
	tr.lpstrText = &strBuff[4];
	SendMessage(m_hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&tr);
	if(strBuff[strBuff.length()-1] == '\0')
		strBuff.erase(strBuff.begin()+strBuff.length()-1);

	// Add open and close format tags
	chFormat = (char)tolower(chFormat);
	strBuff[0] = '<';
	strBuff[1] = '-';
	strBuff[2] = chFormat;
	strBuff[3] = '>';
	strBuff += "<-/";
	strBuff += chFormat;
	strBuff += '>';

	// Set text and reselect selection
	SendMessage(m_hWnd, EM_REPLACESEL, (WPARAM)TRUE, (LPARAM)strBuff.c_str());
	if(bEmpty)
	{
		tr.chrg.cpMin += 4;
		tr.chrg.cpMax = tr.chrg.cpMin;
	}
	else
	{
		tr.chrg.cpMax += 9;
	}
	SendMessage(m_hWnd, EM_EXSETSEL, 0, (LPARAM)&tr.chrg);
}

//============================================================================

static DWORD CALLBACK StreamInCallback(DWORD_PTR dwCookie, LPBYTE pBuff, LONG cb, LONG* pcb)
{
	std::pair<const std::string*, long>& buff = *(std::pair<const std::string*, long>*)dwCookie;
	long nRemain = (long)buff.first->length() - buff.second;
	if(cb > nRemain) cb = nRemain;
	memcpy(pBuff, &buff.first->at(buff.second), cb);
	buff.second += cb;
	*pcb = cb;
	return 0;
}

std::string CDruinkEdit::ConvertRTFString(const std::string& str)
{
	// Set this to m_hWnd for debugging
	HWND hWnd = m_hWndPasteBuffer;

	// Streamin to paste buffer
	EDITSTREAM es;
	std::pair<const std::string*, long> buff(&str, 0);
	es.dwCookie = (DWORD_PTR)&buff;
	es.dwError = 0;
	es.pfnCallback = StreamInCallback;
	SendMessage(hWnd, EM_STREAMIN, SFF_SELECTION|SF_RTF, (LPARAM)&es);

	// Get buffer size
	LONG nLen = (LONG)SendMessage(hWnd, WM_GETTEXTLENGTH, 0, 0);

	// Process chars
	int nCurrFormat = 0;
	COLORREF crCurrent = m_crDefaultText;
	std::string strRet;
	strRet.reserve(nLen);
	for(LONG i=0; i<nLen; ++i)
	{
		// Get this character as text
		TEXTRANGE text;
		char szBuff[2];
		text.chrg.cpMin = i;
		text.chrg.cpMax = text.chrg.cpMin+1;
		text.lpstrText = szBuff;
		if(!SendMessage(hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&text))
			break;

		// Get character format
		CHARFORMAT cfFormat;
		ZeroMemory(&cfFormat, sizeof(cfFormat));
		cfFormat.cbSize = sizeof(cfFormat);
		SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&text.chrg);
		SendMessage(hWnd, EM_GETCHARFORMAT, SCF_SELECTION, (LPARAM)&cfFormat);

		// Convert
		int nNewFormat = 0;
		COLORREF crNew = crCurrent;
		if(cfFormat.dwMask&CFM_BOLD && cfFormat.dwEffects&CFM_BOLD) nNewFormat |= TA2BIT_BOLD;
		if(cfFormat.dwMask&CFM_ITALIC && cfFormat.dwEffects&CFM_ITALIC) nNewFormat |= TA2BIT_ITALIC;
		if(cfFormat.dwMask&CFM_UNDERLINE && cfFormat.dwEffects&CFM_UNDERLINE) nNewFormat |= TA2BIT_UNDERLINE;
		if(cfFormat.dwMask&CFM_STRIKEOUT && cfFormat.dwEffects&CFM_STRIKEOUT) nNewFormat |= TA2BIT_STRIKEOUT;
		if(cfFormat.dwMask&CFM_COLOR) crNew = cfFormat.crTextColor;

		// Don't use black, use default instead
		if(crNew == RGB(0, 0, 0))
			crNew = m_crDefaultText;

		// Handle default text (Reset colour)
		if(crCurrent!=crNew && crNew==m_crDefaultText)
		{
			strRet += "<-0>";
			nCurrFormat = 0;
			crCurrent = crNew;
		}

		// Check for format differences
		if(nCurrFormat != nNewFormat)
		{
			if(nNewFormat == 0)
				strRet += "<-0>";
			else
			{
				if((nNewFormat&TA2BIT_BOLD) && !(nCurrFormat&TA2BIT_BOLD)) strRet += "<-b>";
				else if(!(nNewFormat&TA2BIT_BOLD) && (nCurrFormat&TA2BIT_BOLD)) strRet += "<-/b>";
				if((nNewFormat&TA2BIT_ITALIC) && !(nCurrFormat&TA2BIT_ITALIC)) strRet += "<-i>";
				else if(!(nNewFormat&TA2BIT_ITALIC) && (nCurrFormat&TA2BIT_ITALIC)) strRet += "<-/i>";
				if((nNewFormat&TA2BIT_UNDERLINE) && !(nCurrFormat&TA2BIT_UNDERLINE)) strRet += "<-u>";
				else if(!(nNewFormat&TA2BIT_UNDERLINE) && (nCurrFormat&TA2BIT_UNDERLINE)) strRet += "<-/u>";
				if((nNewFormat&TA2BIT_STRIKEOUT) && !(nCurrFormat&TA2BIT_STRIKEOUT)) strRet += "<-s>";
				else if(!(nNewFormat&TA2BIT_STRIKEOUT) && (nCurrFormat&TA2BIT_STRIKEOUT)) strRet += "<-/s>";
			}
			nCurrFormat = nNewFormat;
		}

		// Check for colour differences
		if(crCurrent != crNew)
		{
			// RichEdit is BGR order, we need RGB
			DWORD dwColour = ((crNew&0x0000ff)<<16) | (crNew&0x00ff00) | ((crNew&0xff0000)>>16);
			char szBuff[32];
			sprintf(szBuff, "<-c%06x>", dwColour);
			strRet += szBuff;
			crCurrent = crNew;
		}

		// Check for special characters
		BYTE byChar = (unsigned char)szBuff[0];
		if(byChar < 128)
		{
			// Newline means formatting is reset
			if(szBuff[0]=='\r' || szBuff[0]=='\n')
			{
				nCurrFormat = 0;
				crCurrent = m_crDefaultText;
			}
			strRet += szBuff[0];
		}
		else
		{
			if(byChar == 0xb7)
				strRet += '*'; // MS word bullet marker
		}
	}

	// Clear paste buffer and return buffer
	SetWindowText(hWnd, "");
	return strRet;
}

//============================================================================

void CDruinkEdit::DoShellExecute(const std::string& strCmd)
{
	m_shellExecuteThread.PostRequest(strCmd);
}

//============================================================================

void CDruinkEdit::DoKeypress(WPARAM nChar, LPARAM nParams)
{
	if(!OnKeyDown((int)nChar, nParams))
		return;

	std::string str("X");
	str[0] = (char)nChar;
	AddTextImpl(str, INT_MAX, false);
}

//============================================================================

void CDruinkEdit::UpdateCursor()
{
	if(m_bMenuVisible)
	{
		SetCursor(m_hArrowCursor);
	}
	else if(m_bOverURL)
	{
		SetCursor(m_hLinkCursor);
	}
	else
	{
		POINT pt;
		RECT rcClient;
		GetCursorPos(&pt);
		ScreenToClient(m_hWnd, &pt);
		GetClientRect(m_hWnd, &rcClient);
		rcClient.right -= rcClient.left;

		if(pt.x >= rcClient.right)
			SetCursor(m_hArrowCursor);
		else
			SetCursor(m_hDefaultCursor);
	}
}

//============================================================================

void CDruinkEdit::OnWindowPosChanged(WINDOWPOS* pNewPos)
{
	// Size changing to zero?
	bool bWasMinimized = m_bMinimized;
	if(!(pNewPos->flags & SWP_NOSIZE) && (pNewPos->cx == 0 || pNewPos->cy == 0))
	{
		m_bMinimized = true;
	}
	else
	{
		m_bMinimized = false;

		// If we were minimized, force an update
		if(bWasMinimized)
			AddText("");
	}
}

//============================================================================

DWORD CDruinkEdit::GetBaseWindowStyle()
{
	return WS_VISIBLE | WS_CHILD | WS_TABSTOP | WS_VSCROLL | ES_AUTOVSCROLL |
		ES_MULTILINE | ES_LEFT | ES_SUNKEN;
}

DWORD CDruinkEdit::GetThisWindowStyle(bool bReadOnly)
{
	if(bReadOnly)
	{
		return GetBaseWindowStyle() | ES_READONLY | ES_SAVESEL | ES_NOHIDESEL |
			ES_DISABLENOSCROLL;
	}
	else
	{
		return GetBaseWindowStyle() | ES_AUTOHSCROLL;
	}
}

//============================================================================

bool CDruinkEdit::IsHTTPUrl(const char* szUrl)
{
	if(strncmp(szUrl, "http://", 7) == 0)
		return true;
	else if(strncmp(szUrl, "www.", 4) == 0)
		return true;
	return false;
}

//============================================================================
