//============================================================================
// ShellExecuteThread.h - Thread to process ShellExecute requests
//============================================================================
#include "ShellExecuteThread.h"

static const DWORD s_dwThreadWaitTime = 5000;

//============================================================================

ShellExecuteThread::ShellExecuteThread() :
	m_hThread(NULL),
	m_hSignal(NULL),
	m_bExit(false)
{
	InitializeCriticalSection(&m_cs);
}

ShellExecuteThread::~ShellExecuteThread()
{
	// Signal thread to exit
	if(m_hThread)
	{
		m_bExit = true;
		SetEvent(m_hSignal);
		if(WaitForSingleObject(m_hThread, s_dwThreadWaitTime) == WAIT_TIMEOUT)
			TerminateThread(m_hThread, 0xffffffff);

		CloseHandle(m_hThread);
		CloseHandle(m_hSignal);
	}

	DeleteCriticalSection(&m_cs);
}

//============================================================================

bool ShellExecuteThread::Init()
{
	// Create signal event
	m_hSignal = CreateEvent(NULL, FALSE, FALSE, NULL);
	if(!m_hSignal)
		return false;

	// Create the thread
	DWORD dwID = 0;
	m_bExit = false;
	m_hThread = CreateThread(NULL, 0, StaticThreadProc, this, 0, &dwID);
	if(!m_hThread)
	{
		CloseHandle(m_hSignal);
		m_hSignal = NULL;
		return false;
	}

	// Done
	return true;
}

void ShellExecuteThread::PostRequest(const std::string& str)
{
	EnterCriticalSection(&m_cs);
	m_strRequest = str;
	LeaveCriticalSection(&m_cs);
	SetEvent(m_hSignal);
}

//============================================================================

DWORD WINAPI ShellExecuteThread::StaticThreadProc(LPVOID pParam)
{
	ShellExecuteThread* pThread = (ShellExecuteThread*)pParam;
	return pThread->ThreadProc();
}

DWORD ShellExecuteThread::ThreadProc()
{
	while(!m_bExit)
	{
		// Wait for signal and copy request
		WaitForSingleObject(m_hSignal, INFINITE);
		EnterCriticalSection(&m_cs);
		std::string strRequest = m_strRequest;
		m_strRequest = "";
		LeaveCriticalSection(&m_cs);

		// Handle request
		if(!strRequest.empty())
			ShellExecute(NULL, "open", strRequest.c_str(), NULL, NULL, SW_SHOWNORMAL);
	}

	return 0;
}

//============================================================================
