//============================================================================
// HTTPSocket.h - Socket that communicates over psuedo-HTTP
//============================================================================

#ifndef __HTTPSOCKET_H__
#define __HTTPSOCKET_H__

#include "BaseSocket/BaseSocket.h"
#include "Socket/Socket.h"

class CHTTPSocket : public CBaseSocket
{
public:
	CHTTPSocket();
	virtual ~CHTTPSocket();
	virtual void Destroy();

	// Perform a nonblocking connect
	virtual bool StartConnect(const char* szHost, unsigned short wPort);
	virtual ConnectStatus PollConnectComplete();

	// Disconnect and cleanup the socket
	virtual void Disconnect();

	// Reset error string, and nullify connect/socket error stats
	virtual void ClearError();

	// Check the socket is still connected, and read data if possible
	virtual SocketStatus Poll();

	// Get one line from the buffer
	virtual const char* GetLine();

	// Get the whole buffer
	virtual size_t GetBufferSize() const { return m_buffer.size(); }
	virtual const char* GetBuffer() const { return m_buffer.empty() ? "" : &m_buffer[0]; }

	// Clear the whole buffer
	virtual void ClearBuffer() { m_buffer.clear(); }

	// Send a buffer
	virtual bool Send(const void* buff, size_t len);

	// Accessors
	virtual const char* GetError() const { return m_error.c_str(); }
	virtual bool IsConnected() const { return m_connectStatus == STATUS_Connected; }
	virtual bool DataPending() const { return !m_buffer.empty(); }

protected:
	enum InternalState
	{
		STATE_Idle,					// Doing nothing
		STATE_ConnectingHTTP,		// Doing an HTTP connection
		STATE_WaitResponse,			// Waiting on server replying to our header
	};

	void Tick();
	void TickConnecting();
	void TickWaitResponse();

	CSocket m_sock;
	std::vector<char> m_buffer;			// Buffer for data from telnet server
	std::vector<char> m_sendBuffer;		// Outgoing data buffer
	InternalState m_state;
	ConnectStatus m_connectStatus;		// Status of virtual connection
	SocketStatus m_socketStatus;
	std::string m_lineBuff;
	std::string m_error;
	std::string m_cookie;

	std::string m_host;
	unsigned short m_port;
};

#endif // __HTTPSOCKET_H__
