//=================================================================================================
// Memory.h - Custom memory manager
//=================================================================================================

#ifndef __MEMORY_H__
#define __MEMORY_H__

#include <windows.h>

#ifdef _DEBUG
#	define USE_MMGR
#endif

//=================================================================================================

#ifdef USE_MMGR

#define USE_STACKTRACE

class Memory
{
public:
	Memory();
	~Memory();

	// Used internally:
	static Memory& Get()
	{
		return *(Memory*)ms_pInstance;
	}
	static void Create();
	static void Destroy();

	//=============================================================================================

	void DumpAllAllocations(bool bDumpToFile);
	size_t GetNextAllocID() const { return m_nNextID; }
	size_t GetNumActiveAllocations() const { return m_nAllocations; }
	size_t GetAllocatedBytes() const { return m_nAllocatedBytes; }
	size_t GetHighWatermark() const { return m_nHighWatermark; }
	void BeginAllocationCount() { m_nAllocationsSinceMarker = 0; }
	size_t GetAllocationCount() const { return m_nAllocationsSinceMarker; }
	void SetAllocationTagID(size_t nTag) { m_nAllocationTag = nTag; }
	void SetBreakOnAllocID(size_t nID) { m_nBreakOnAllocID = nID; }

	//=============================================================================================

	void* Allocate(size_t nSize);
	void* AllocateArray(size_t nSize);
	void Deallocate(void* pAddress);
	void DeallocateArray(void* pAddress);

private:
	enum Alloc
	{
		ALLOC_New,
		ALLOC_NewArray,
	};

	struct AllocHeader
	{
		AllocHeader* pNext;
		AllocHeader* pPrev;
		Alloc eType;
		size_t nLength;
		size_t nAllocID;
		size_t nTagID;
		#ifdef USE_STACKTRACE
			static const size_t cnBufferSize = 256;
			char szFile[cnBufferSize];
			unsigned int nLine;
			char szFunc[cnBufferSize];
		#endif
	};

	typedef void (WINAPI *LPRtlCaptureContext)(CONTEXT* ContextRecord);

	void* InternalAlloc(size_t nSize, Alloc eType);
	void InternalDealloc(void* p, Alloc eType);
	void DetermineCaller(AllocHeader* pAllocation);

	static Memory*		ms_pInstance;

	LPRtlCaptureContext	m_pfnCaptureContext;
	CRITICAL_SECTION	m_theMutex;
	AllocHeader*		m_pHead;
	AllocHeader*		m_pTail;
	size_t				m_nAllocations;
	size_t				m_nAllocationsSinceMarker;
	size_t				m_nAllocatedBytes;
	size_t				m_nNextID;
	size_t				m_nBreakOnAllocID;
	size_t				m_nHighWatermark;
	size_t				m_nAllocationTag;
};

#endif // USE_MMGR
#endif // __MEMORY_H__
