//============================================================================
// ScriptVar.h - A single script variable
//============================================================================

#ifndef __SCRIPTVAR_H__
#define __SCRIPTVAR_H__

class ScriptVar
{
public:
	enum Type
	{
		TYPE_Float,
		TYPE_Int,
		TYPE_String,
		TYPE_Array
	};

	//========================================================================

	ScriptVar();
	ScriptVar(Type eType); // Initialise to empty
	ScriptVar(float f);
	ScriptVar(int n);
	ScriptVar(const char* sz);
	ScriptVar(const ScriptVar* pArray, size_t nLength);
	ScriptVar(const ScriptVar& rhs);
	~ScriptVar();

	//========================================================================

	// Get type
	Type GetType() const { return m_eType; }

	// Coerce to a certain type
	int ToInt() const;
	float ToFloat() const;
	const char* ToString() const;	// Uses static buffer, only valid for one call

	// Array functions. Indices are 0-based
	size_t GetArraySize() const { return m_nArrayLen; }
	const ScriptVar& GetArrayItem(size_t nIdx) const;
	ScriptVar& GetArrayItem(size_t nIdx);
	void AddArrayItem(const ScriptVar& val);
	void RemoveArrayItem(size_t nIdx);

	// Operators
	ScriptVar& operator[](size_t nIdx) { return GetArrayItem(nIdx); }
	const ScriptVar& operator[](size_t nIdx) const { return GetArrayItem(nIdx); }
	ScriptVar& operator=(const ScriptVar& rhs);
	ScriptVar& operator=(float f);
	ScriptVar& operator=(int n);
	ScriptVar& operator=(const char* sz);
	ScriptVar& AssignToArray(const ScriptVar* pArray, size_t nLength);
	bool operator==(const ScriptVar& rhs) const;
	bool operator!=(const ScriptVar& rhs) const { return !(*this == rhs); }

private:
	union UData
	{
		float fValue;
		int nValue;
		char* szValue;
		ScriptVar* pValue;
	};

	typedef void (*LPfree)(void* p);

	Type m_eType;

	UData m_primVal;
	size_t m_nArrayLen;
	LPfree m_pfnFree;
};

#endif // __SCRIPTVAR_H__
