//============================================================================
// ScriptContextInterface.h - Interface implemented by script DLLs
//============================================================================

#ifndef __SCRIPTCONTEXTINTERFACE_H__
#define __SCRIPTCONTEXTINTERFACE_H__

#include "Plugin/ScriptContext.h"

//============================================================================

enum ScriptLoadResult
{
	SLR_Success,
	SLR_AlreadyLoaded,
	SLR_FileNotFound,
	SLR_ScriptError,
	SLR_UnknownError
};

//============================================================================

class ScriptContextInterface : public ScriptContext
{
public:
	typedef bool (*LPScriptFunction)(ScriptContext& context, void* pParam);

	ScriptContextInterface() {}
	virtual ~ScriptContextInterface() {}

	//========================================================================
	// ScriptContext:
	virtual size_t GetNumArgs() const = 0;
	virtual ScriptVar GetArg(size_t nIdx) const = 0;
	virtual void AddRetval(const ScriptVar& var) = 0;
	virtual void AddArg(const ScriptVar& var) = 0;
	virtual void SetError(const char* szError) = 0;
	virtual const char* GetScriptType() = 0;

	//========================================================================
	// ScriptContextInterface:

	// Destroy this context (Usually call delete this)
	virtual void Destroy() = 0;

	// Load and run a script from a file on disk
	virtual ScriptLoadResult LoadScript(const char* szScriptName) = 0;

	// Load and run a script contained in a data buffer
	virtual bool LoadScript(const void* pData, size_t nLen) = 0;

	// Reload all scripts
	virtual bool ReloadScripts() = 0;

	// Register / unregister a script-callable C function
	virtual bool Register(LPScriptFunction pfnFunc, void* pParam, const char* szName) = 0;
	virtual bool Unregister(const char* szFuncName) = 0;

	// Call a function contained in this script. On successful completion, ppRetvals
	// will contain the return values, and rnRetvals will contain the size of the
	// ppRetvals array.
	virtual bool Call(const char* szFunction, size_t nArgs, const ScriptVar* pArgs,
		ScriptVar** ppRetvals, size_t& rnRetvals) = 0;

	// Return the error string
	virtual const char* GetError() const = 0;
};

//============================================================================
// Exported functions
#ifdef BUILDING_SCRIPT_DLL
#	define SCRIPT_LINKAGE __declspec(dllexport)
#else
#	define SCRIPT_LINKAGE __declspec(dllimport)
#endif

#define SCRIPT_API_VERSION 1

// Exported function
extern "C"
{
	SCRIPT_LINKAGE ScriptContextInterface* CreateScriptInterface(int nAPIVersion);
};
//============================================================================

#endif // __SCRIPTCONTEXTINTERFACE_H__
