//============================================================================
// Socket.cpp - Basic socket class
//============================================================================
#include <winsock2.h>
#include "Plugin/Assert.h"
#include "Socket.h"
#include <sstream>

#pragma comment (lib,"ws2_32.lib")

#define FORCE_SLOW_SOCKET 0
#define SLOW_SOCKET_MAX_BUFF_PER_POLL 256

//============================================================================
class CInit
{
public:
	CInit()
	{
		WSADATA wsaData;
		WSAStartup(MAKEWORD(1,1),&wsaData);
	}

	~CInit()
	{
		WSACleanup();
	}
};
CInit g_init;


//============================================================================
static std::string WSAErrorAsString(int nError)
{
std::stringstream strRet;

	strRet << "WinSock Error Code " << nError << ": ";

	switch(nError)
	{
	case WSAEACCES: strRet << "Access denied."; break;
	case WSAEADDRINUSE: strRet << "Address already in use."; break;
	case WSAEADDRNOTAVAIL: strRet << "Address not availiable."; break;
	case WSAEAFNOSUPPORT: strRet << "Address family not supported."; break;
	case WSAEALREADY: strRet << "Operation already in progress."; break;
	case WSAECONNABORTED: strRet << "Software caused connection abort."; break;
	case WSAECONNREFUSED: strRet << "Connection refused by remote machine. Remote server is most likely offline."; break;
	case WSAECONNRESET: strRet << "Connection reset by peer. Remote server forcibly closed connection. Likely cause is computer reboot, or software crash on the server."; break;
	case WSAEDESTADDRREQ: strRet << "Destination address required."; break;
	case WSAEFAULT: strRet << "Internal software error."; break;
	case WSAEHOSTDOWN: strRet << "Host is down."; break;
	case WSAEHOSTUNREACH: strRet << "Host unreachable. An operation attempted to an unreachable host."; break;
	case WSAEINPROGRESS: strRet << "Operation now in progress."; break;
	case WSAEINTR: strRet << "Interrupted function call."; break;
	case WSAEINVAL: strRet << "Invalid value."; break;
	case WSAEISCONN: strRet << "Already connected."; break;
	case WSAEMFILE: strRet << "Too many open sockets."; break;
	case WSAEMSGSIZE: strRet << "Message too long."; break;
	case WSAENETDOWN: strRet << "Network is down. An operation encountered a dead network. This could indicate a serious failure of the internal network system, the network interface, or the local network itself."; break;
	case WSAENETRESET: strRet << "Network dropped connection on reset. The connection has been broken due to \"keep-alive\" activity detecting a failure while the operation was in progress."; break;
	case WSAENETUNREACH: strRet << "Network is unreachable. An operation was attempted to an unreachable network. This usually means the local software knows no route to reach the remote host."; break;
	case WSAENOBUFS: strRet << "No buffer space availiable."; break;
	case WSAENOPROTOOPT: strRet << "Internal error: bad protocol option."; break;
	case WSAENOTCONN: strRet << "Failed to send or receive data because the socket is not connected."; break;
	case WSAENOTSOCK: strRet << "Internal error: A function was called on a socket handle that was invalid."; break;
	case WSAEOPNOTSUPP: strRet << "Operation not supported."; break;
	case WSAEPROCLIM: strRet << "Too many processes accessing the WinSock library."; break;
	case WSAEPROTONOSUPPORT: strRet << "Protocol not supported."; break;
	case WSAEPROTOTYPE: strRet << "Internal error: protocol wrong type for socket."; break;
	case WSAESHUTDOWN: strRet << "Data send was attempted after socket has been shutdown."; break;
	case WSAESOCKTNOSUPPORT: strRet << "Socket type not supported."; break;
	case WSAETIMEDOUT: strRet << "Connection timed out."; break;
	case WSAEWOULDBLOCK: strRet << "Internal error: Resource temporarily unavailable."; break;
	case WSANOTINITIALISED: strRet << "Internal error: Socket library not initialised."; break;
	case WSAVERNOTSUPPORTED: strRet << "Version not supported. Your WinSock version is inadequate for this application."; break;
	case WSAEDISCON: strRet << "Graceful shutdown in progress."; break;
	case WSAEPFNOSUPPORT: strRet << "Protocol family not supported."; break;
	case WSAETOOMANYREFS: strRet << "Too many references."; break;
	case WSAELOOP: strRet << "WSAELOOP"; break;
	case WSAENAMETOOLONG: strRet << "Name too long."; break;
	case WSAENOTEMPTY: strRet << "WSAENOTEMPTY"; break;
	case WSAEUSERS: strRet << "WSAEUSERS"; break;
	case WSAEDQUOT: strRet << "WSAEDQUOT"; break;
	case WSAESTALE: strRet << "WSAESTALE"; break;
	case WSAEREMOTE: strRet << "WSAEREMOTE"; break;
	case WSASYSNOTREADY: strRet << "Network subsystem is unavailable."; break;
	case WSAENOMORE: strRet << "WSAENOMORE"; break;
	case WSAECANCELLED: strRet << "WSAECANCELLED"; break;
	case WSAEINVALIDPROCTABLE: strRet << "Invalid procedure table from service provider."; break;
	case WSAEINVALIDPROVIDER: strRet << "Invalid service provider version number."; break;
	case WSAEPROVIDERFAILEDINIT: strRet << "Unable to initialize a service provider."; break;
	case WSASYSCALLFAILURE: strRet << "System call failure."; break;
	case WSASERVICE_NOT_FOUND: strRet << "WSASERVICE_NOT_FOUND"; break;
	case WSATYPE_NOT_FOUND: strRet << "Class type not found."; break;
	case WSA_E_NO_MORE: strRet << "WSA_E_NO_MORE"; break;
	case WSA_E_CANCELLED: strRet << "WSA_E_CANCELLED"; break;
	case WSAEREFUSED: strRet << "WSAEREFUSED"; break;
	case WSAHOST_NOT_FOUND: strRet << "Host not found."; break;
	case WSA_INVALID_HANDLE: strRet << "Specified event object handle is invalid."; break;
	case WSA_INVALID_PARAMETER: strRet << "One or more parameters are invalid."; break;
	case WSA_IO_INCOMPLETE: strRet << "Overlapped I/O event object not in signaled state."; break;
	case WSA_IO_PENDING: strRet << "Overlapped operations will complete later."; break;
	case WSA_NOT_ENOUGH_MEMORY: strRet << "Insufficient memory available."; break;
	case WSANO_DATA: strRet << "Valid name, no data record of requested type."; break;
	case WSANO_RECOVERY: strRet << "This is a non-recoverable error."; break;
	case WSATRY_AGAIN: strRet << "Non-authoritative host not found."; break;
	case WSA_OPERATION_ABORTED : strRet << "Overlapped operation aborted."; break;
	default: strRet << "Unknown error."; break;
	}

	return strRet.str();
}

//============================================================================
CSocket::CSocket() :
	m_sock(INVALID_SOCKET),
	m_eConnectStatus(STATUS_NotConnected),
	m_eSocketStatus(STATUS_Ok),
	m_feedingBuffer(false)
{
}

CSocket::~CSocket()
{
}

void CSocket::Destroy()
{
	delete this;
}

//============================================================================
bool CSocket::StartConnect(const char* szHost, unsigned short wPort)
{
	// Disconnect if connected
	Disconnect();

	// Create Socket
	m_sock = socket(PF_INET, SOCK_STREAM, 0);
	if(m_sock == INVALID_SOCKET)
	{
		m_strError = "socket() Failed! " + WSAErrorAsString(WSAGetLastError());
		return false;
	}

	// Setup Address
	sockaddr_in sockAddr;
	memset(&sockAddr, 0, sizeof(sockAddr));
	sockAddr.sin_addr.s_addr = inet_addr(szHost);
	if(sockAddr.sin_addr.s_addr == INADDR_NONE)
	{
		hostent* pHost = gethostbyname(szHost);
		if(!pHost)
		{
			closesocket(m_sock);
			m_sock = INVALID_SOCKET;
			m_strError = "Could not look up IP address for host ";
			m_strError += szHost;
			return false;
		}
		sockAddr.sin_addr = *(in_addr*)pHost->h_addr;
	}

	// Setup rest of address
	sockAddr.sin_family = AF_INET;
	sockAddr.sin_port = htons(wPort);

	// Set socket as nonblocking
	unsigned long nNonblocking = 1;
	ioctlsocket(m_sock, FIONBIO, &nNonblocking);

	// Begin connect
	if(connect(m_sock, (sockaddr*)&sockAddr, sizeof(sockaddr_in)) == SOCKET_ERROR)
	{
		int nError = WSAGetLastError();
		if(nError != WSAEWOULDBLOCK)
		{
			m_strError = "connect() Failed! " + WSAErrorAsString(WSAGetLastError());
			closesocket(m_sock);
			m_sock = INVALID_SOCKET;
			return false;
		}
		else
		{
			// Nonblocking connect in progress
			m_eConnectStatus = STATUS_Connecting;
			return true;
		}
	}

	// Connected instantly
	m_eConnectStatus = STATUS_Connected;
	return true;
}

#pragma warning(disable:4127)	// warning C4127: conditional expression is constant
CBaseSocket::ConnectStatus CSocket::PollConnectComplete()
{
	// Early out if we can
	if(m_eConnectStatus != STATUS_Connecting)
		return m_eConnectStatus;

	// Setup write and connect sets, and a zero timeout
	fd_set theWriteSet, theErrorSet;
	FD_ZERO(&theWriteSet);
	FD_SET(m_sock, &theWriteSet);
	FD_ZERO(&theErrorSet);
	FD_SET(m_sock, &theErrorSet);
	timeval theTimeout;
	theTimeout.tv_sec = 0;
	theTimeout.tv_usec = 0;

	// Poll socket
	int nRet = select(0, NULL, &theWriteSet, &theErrorSet, &theTimeout);
	if(nRet == 0)
	{
		// Timed out
		return STATUS_Connecting;
	}
	else if(nRet == SOCKET_ERROR)
	{
		m_strError = "select() Failed! " + WSAErrorAsString(WSAGetLastError());
		closesocket(m_sock);
		m_sock = INVALID_SOCKET;
		m_eConnectStatus = STATUS_ConnectError;
		return m_eConnectStatus;
	}

	// Is it a successful connect?
	if(FD_ISSET(m_sock, &theWriteSet))
	{
		// Socket is writeable => Connected ok
		m_eConnectStatus = STATUS_Connected;
		return m_eConnectStatus;
	}

	// Error must be set
	int nError = 0;
	int nLen = sizeof(nError);
	getsockopt(m_sock, SOL_SOCKET, SO_ERROR, (char*)&nError, &nLen);
	m_strError = "connect() Failed! " + WSAErrorAsString(nError);
	closesocket(m_sock);
	m_sock = INVALID_SOCKET;
	m_eConnectStatus = STATUS_ConnectError;
	return m_eConnectStatus;
}
#pragma warning(default:4127)	// warning C4127: conditional expression is constant

void CSocket::Disconnect()
{
	if(m_sock != INVALID_SOCKET)
	{
		shutdown(m_sock, SD_BOTH);
		closesocket(m_sock);
		m_sock = INVALID_SOCKET;
	}

	m_strError = "";
	m_eConnectStatus = STATUS_NotConnected;
	m_eSocketStatus = STATUS_Ok;

	m_vBuffer.clear();
}

void CSocket::ClearError()
{
	m_strError = "";
	if(m_eConnectStatus == STATUS_ConnectError)
		m_eConnectStatus = STATUS_NotConnected;
	m_eSocketStatus = STATUS_Ok;
}

//============================================================================
#pragma warning(disable:4127)	// warning C4127: conditional expression is constant
CBaseSocket::SocketStatus CSocket::Poll()
{
	// Early out if we can
	#if FORCE_SLOW_SOCKET
		if(m_eSocketStatus == STATUS_Disconnected && m_feedingBuffer)
			return STATUS_Ok;
	#endif
	if(m_eConnectStatus != STATUS_Connected)
		return STATUS_Disconnected;
	if(m_eSocketStatus != STATUS_Ok)
		return m_eSocketStatus;

	// Setup read and error sets, and a zero timeout
	fd_set theReadSet, theErrorSet;
	FD_ZERO(&theReadSet);
	FD_SET(m_sock, &theReadSet);
	FD_ZERO(&theErrorSet);
	FD_SET(m_sock, &theErrorSet);
	timeval theTimeout;
	theTimeout.tv_sec = 0;
	theTimeout.tv_usec = 0;

	// Poll socket
	int nRet = select(0, &theReadSet, NULL, &theErrorSet, &theTimeout);
	if(nRet == 0)
	{
		// Timed out. Any socket error set?
		int error = 0;
		int len = sizeof(error);
		nRet = getsockopt(m_sock, SOL_SOCKET, SO_ERROR, (char*)&error, &len);
		if(nRet != 0 || error != 0)
		{
			if(nRet != 0)
			{
				m_strError = "getsockopt() Failed! " + WSAErrorAsString(WSAGetLastError());
			}
			else
			{
				m_strError = "getsockopt() reported an error, though select() did not! " +
					WSAErrorAsString(WSAGetLastError());
			}
			closesocket(m_sock);
			m_sock = INVALID_SOCKET;
			m_eConnectStatus = STATUS_NotConnected;
			m_eSocketStatus = STATUS_Error;
		}
	}

	// Error from select()?
	else if(nRet == SOCKET_ERROR)
	{
		m_strError = "select() Failed! " + WSAErrorAsString(WSAGetLastError());
		closesocket(m_sock);
		m_sock = INVALID_SOCKET;
		m_eConnectStatus = STATUS_NotConnected;
		m_eSocketStatus = STATUS_Error;
	}

	// Is it an error?
	else if(FD_ISSET(m_sock, &theErrorSet))
	{
		int nError = 0;
		int nLen = sizeof(nError);
		getsockopt(m_sock, SOL_SOCKET, SO_ERROR, (char*)&nError, &nLen);
		m_strError = "select() error is set: " + WSAErrorAsString(nError);
		closesocket(m_sock);
		m_sock = INVALID_SOCKET;
		m_eConnectStatus = STATUS_NotConnected;
		m_eSocketStatus = STATUS_Error;
	}

	// Must have been a read
	else
	{
		char szBuff[256];
		int nRet = 0;
		do {
			// Try to read
			nRet = recv(m_sock, szBuff, sizeof(szBuff), 0);
			if(nRet == SOCKET_ERROR)
			{
				// If the socket has a multiple of sizeof(szBuff) bytes in it, then the
				// next loop iteration it'll give WSAEWOULDBLOCK
				int nError = WSAGetLastError();
				if(nError != WSAEWOULDBLOCK)
				{
					// Error
					m_strError = "recv() Failed! " + WSAErrorAsString(nError);
					closesocket(m_sock);
					m_sock = INVALID_SOCKET;
					m_eConnectStatus = STATUS_NotConnected;
					m_eSocketStatus = STATUS_Error;
				}
			}
			else if(nRet == 0)
			{
				// Clean disconnect
				closesocket(m_sock);
				m_sock = INVALID_SOCKET;
				m_eConnectStatus = STATUS_NotConnected;
				m_eSocketStatus = STATUS_Disconnected;
			}
			else
			{
				// Some data pending
				size_t nOldLen = m_vBuffer.size();
				m_vBuffer.resize(nOldLen + (size_t)nRet);
				memcpy(&m_vBuffer[nOldLen], szBuff, nRet);
			}
		} while(nRet == sizeof(szBuff));
	}

	#if FORCE_SLOW_SOCKET
		if(m_eSocketStatus == STATUS_Disconnected && m_feedingBuffer)
			return STATUS_Ok;
	#endif

	return m_eSocketStatus;
}

#pragma warning(default:4127)	// warning C4127: conditional expression is constant

//============================================================================

const char* CSocket::GetLine()
{
	for(size_t i=0; i<m_vBuffer.size(); ++i)
	{
		if(m_vBuffer[i] == '\n')
		{
			m_lineBuff.resize(i+1, '*');
			memcpy(&m_lineBuff[0], &m_vBuffer[0], i+1);
			m_vBuffer.erase(m_vBuffer.begin(), m_vBuffer.begin()+i+1);
			return m_lineBuff.c_str();
		}
	}

	return "";
}

size_t CSocket::GetBufferSize() const
{
	#if FORCE_SLOW_SOCKET
		if(m_vBuffer.size() > SLOW_SOCKET_MAX_BUFF_PER_POLL)
		{
			m_feedingBuffer = true;
			return SLOW_SOCKET_MAX_BUFF_PER_POLL;
		}
	#endif

	m_feedingBuffer = false;
	return m_vBuffer.size();
}

const char* CSocket::GetBuffer() const
{
	return m_vBuffer.empty() ? "" : &m_vBuffer[0];
}

void CSocket::ClearBuffer()
{
	#if FORCE_SLOW_SOCKET
		if(m_vBuffer.size() > SLOW_SOCKET_MAX_BUFF_PER_POLL)
			m_vBuffer.erase(m_vBuffer.begin(), m_vBuffer.begin()+SLOW_SOCKET_MAX_BUFF_PER_POLL);
		else
			m_vBuffer.clear();
	#else
		m_vBuffer.clear();
	#endif
}

//============================================================================
bool CSocket::Send(const void* buff, size_t len)
{
	int nRet = send(m_sock, (char*)buff, (int)len, 0);
	return nRet == (int)len;
}
