//============================================================================
// Socket.h - Basic socket class
//============================================================================

#ifndef __SOCKET_H__
#define __SOCKET_H__

#include "BaseSocket/BaseSocket.h"
#include <windows.h>
#include <string>
#include <vector>

//============================================================================
class CSocket : public CBaseSocket
{
public:
	CSocket();
	virtual ~CSocket();
	virtual void Destroy();

	// Perform a nonblocking connect
	virtual bool StartConnect(const char* szHost, unsigned short wPort);
	virtual ConnectStatus PollConnectComplete();

	// Disconnect and cleanup the socket
	virtual void Disconnect();

	// Reset error string, and nullify connect/socket error stats
	virtual void ClearError();

	// Check the socket is still connected, and read data if possible
	virtual SocketStatus Poll();

	// Get one line from the buffer
	virtual const char* GetLine();

	// Get the whole buffer
	virtual size_t GetBufferSize() const;
	virtual const char* GetBuffer() const;

	// Clear the whole buffer
	virtual void ClearBuffer();

	// Send a buffer
	virtual bool Send(const void* buff, size_t len);

	// Accessors
	virtual const char* GetError() const { return m_strError.c_str(); }
	virtual bool IsConnected() const { return m_eConnectStatus == STATUS_Connected; }
	virtual bool DataPending() const { return !m_vBuffer.empty(); }

protected:
	UINT_PTR			m_sock;
	std::string			m_strError;
	ConnectStatus		m_eConnectStatus;
	SocketStatus		m_eSocketStatus;
	std::vector<char>	m_vBuffer;
	std::string			m_lineBuff;
	mutable bool		m_feedingBuffer;
};

#endif // __SOCKET_H__
