//============================================================================
// GetOS.cpp - Standard library function: getOS()
//============================================================================
#include "Plugin/ScriptContext.h"
#include <windows.h>
#include <string.h>
#include <string>

// GetProductInfo() related stuff
// typedef'd here to avoid the DLL compiling to need Vista+
typedef BOOL (WINAPI *LPGetProductInfo)(DWORD, DWORD, DWORD, DWORD, PDWORD);
#define PRODUCT_UNDEFINED							0x00000000
#define PRODUCT_ULTIMATE							0x00000001
#define PRODUCT_HOME_BASIC							0x00000002
#define PRODUCT_HOME_PREMIUM						0x00000003
#define PRODUCT_ENTERPRISE							0x00000004
#define	PRODUCT_HOME_BASIC_N						0x00000005
#define PRODUCT_BUSINESS							0x00000006
#define PRODUCT_STANDARD_SERVER						0x00000007
#define PRODUCT_DATACENTER_SERVER					0x00000008
#define PRODUCT_SMALLBUSINESS_SERVER				0x00000009
#define PRODUCT_ENTERPRISE_SERVER					0x0000000A
#define PRODUCT_STARTER								0x0000000B
#define PRODUCT_DATACENTER_SERVER_CORE				0x0000000C
#define PRODUCT_STANDARD_SERVER_CORE				0x0000000D
#define PRODUCT_ENTERPRISE_SERVER_CORE				0x0000000E
#define PRODUCT_ENTERPRISE_SERVER_IA64				0x0000000F
#define PRODUCT_BUSINESS_N							0x00000010
#define PRODUCT_WEB_SERVER							0x00000011
#define PRODUCT_CLUSTER_SERVER						0x00000012
#define PRODUCT_STORAGE_EXPRESS_SERVER				0x00000014
#define PRODUCT_STORAGE_STANDARD_SERVER				0x00000015
#define PRODUCT_STORAGE_WORKGROUP_SERVER			0x00000016
#define PRODUCT_STORAGE_ENTERPRISE_SERVER			0x00000017
#define PRODUCT_SERVER_FOR_SMALLBUSINESS			0x00000018
#define PRODUCT_SMALLBUSINESS_SERVER_PREMIUM		0x00000019
#define PRODUCT_HOME_PREMIUM_N						0x0000001A
#define PRODUCT_ENTERPRISE_N						0x0000001B
#define PRODUCT_ULTIMATE_N							0x0000001C
#define PRODUCT_WEB_SERVER_CORE						0x0000001D
#define PRODUCT_MEDIUMBUSINESS_SERVER_MANAGEMENT	0x0000001E
#define PRODUCT_MEDIUMBUSINESS_SERVER_SECURITY		0x0000001F
#define PRODUCT_MEDIUMBUSINESS_SERVER_MESSAGING		0x00000020
#define PRODUCT_SERVER_FOUNDATION					0x00000021
#define PRODUCT_SERVER_FOR_SMALLBUSINESS_V			0x00000023
#define PRODUCT_STANDARD_SERVER_V					0x00000024
#define PRODUCT_DATACENTER_SERVER_V					0x00000025
#define PRODUCT_ENTERPRISE_SERVER_V					0x00000026
#define PRODUCT_DATACENTER_SERVER_CORE_V			0x00000027
#define PRODUCT_STANDARD_SERVER_CORE_V				0x00000028
#define PRODUCT_ENTERPRISE_SERVER_CORE_V			0x00000029
#define PRODUCT_HYPERV								0x0000002A
#define PRODUCT_STARTER_N							0x0000002F
#define PRODUCT_PROFESSIONAL						0x00000030
#define PRODUCT_PROFESSIONAL_N						0x00000031
#define PRODUCT_STARTER_E							0x00000042
#define PRODUCT_HOME_BASIC_E						0x00000043
#define PRODUCT_HOME_PREMIUM_E						0x00000044
#define PRODUCT_PROFESSIONAL_E						0x00000045
#define PRODUCT_ENTERPRISE_E						0x00000046
#define PRODUCT_ULTIMATE_E							0x00000047
#define PRODUCT_UNLICENSED							0xABCDABCD
#define SM_SERVERR2									89

//============================================================================

std::string GetProductType(DWORD dwType)
{
	switch(dwType)
	{
	case PRODUCT_UNDEFINED: return "(Unknown product type)";
	case PRODUCT_ULTIMATE: return "Ultimate";
	case PRODUCT_HOME_BASIC: return "Home Basic";
	case PRODUCT_HOME_PREMIUM: return "Home Premium";
	case PRODUCT_ENTERPRISE: return "Enterprise";
	case PRODUCT_HOME_BASIC_N: return "Home Basic N";
	case PRODUCT_BUSINESS: return "Business";
	case PRODUCT_STANDARD_SERVER: return "Server Standard";
	case PRODUCT_DATACENTER_SERVER: return "Server Datacenter";
	case PRODUCT_SMALLBUSINESS_SERVER: return "Windows Small Business Server";
	case PRODUCT_ENTERPRISE_SERVER: return "Server Enterprise";
	case PRODUCT_STARTER: return "Starter";
	case PRODUCT_DATACENTER_SERVER_CORE: return "Server Datacenter (core installation)";
	case PRODUCT_STANDARD_SERVER_CORE: return "Server Standard (core installation)";
	case PRODUCT_ENTERPRISE_SERVER_CORE: return "Server Enterprise (core installation)";
	case PRODUCT_ENTERPRISE_SERVER_IA64: return "Server Enterprise for Itanium-based Systems";
	case PRODUCT_BUSINESS_N: return "Business N";
	case PRODUCT_WEB_SERVER: return "Web Server";
	case PRODUCT_CLUSTER_SERVER: return "HPC Edition";
	case PRODUCT_STORAGE_EXPRESS_SERVER: return "Storage Server Express";
	case PRODUCT_STORAGE_STANDARD_SERVER: return "Storage Server Standard";
	case PRODUCT_STORAGE_WORKGROUP_SERVER: return "Storage Server Workgroup";
	case PRODUCT_STORAGE_ENTERPRISE_SERVER: return "Storage Server Enterprise";
	case PRODUCT_SERVER_FOR_SMALLBUSINESS: return "Windows Server 2008 for Windows Essential Server Solutions";
	case PRODUCT_SMALLBUSINESS_SERVER_PREMIUM: return "Small Business Server Premium";
	case PRODUCT_HOME_PREMIUM_N: return "Home Premium N";
	case PRODUCT_ENTERPRISE_N: return "Enterprise N";
	case PRODUCT_ULTIMATE_N: return "Ultimate N";
	case PRODUCT_WEB_SERVER_CORE: return "Web Server (core installation)";
	case PRODUCT_MEDIUMBUSINESS_SERVER_MANAGEMENT: return "Windows Essential Business Server Management Server";
	case PRODUCT_MEDIUMBUSINESS_SERVER_SECURITY: return "Windows Essential Business Server Security Server";
	case PRODUCT_MEDIUMBUSINESS_SERVER_MESSAGING: return "Windows Essential Business Server Messaging Server";
	case PRODUCT_SERVER_FOUNDATION: return "Server Foundation";
	case PRODUCT_SERVER_FOR_SMALLBUSINESS_V: return "Windows Server 2008 without Hyper-V for Windows Essential Server Solutions";
	case PRODUCT_STANDARD_SERVER_V: return "Server Standard without Hyper-V ";
	case PRODUCT_DATACENTER_SERVER_V: return "Server Datacenter without Hyper-V";
	case PRODUCT_ENTERPRISE_SERVER_V: return "Server Enterprise without Hyper-V";
	case PRODUCT_DATACENTER_SERVER_CORE_V: return "Server Datacenter without Hyper-V (core installation)";
	case PRODUCT_STANDARD_SERVER_CORE_V: return "Server Standard without Hyper-V (core installation)";
	case PRODUCT_ENTERPRISE_SERVER_CORE_V: return "Server Enterprise without Hyper-V (core installation)";
	case PRODUCT_HYPERV: return "Microsoft Hyper-V Server";
	case PRODUCT_STARTER_N: return "Starter N";
	case PRODUCT_PROFESSIONAL: return "Professional";
	case PRODUCT_PROFESSIONAL_N: return "Professional N";
	case PRODUCT_STARTER_E: return "Starter E";
	case PRODUCT_HOME_BASIC_E: return "Home Basic E";
	case PRODUCT_HOME_PREMIUM_E: return "Home Premium E";
	case PRODUCT_PROFESSIONAL_E: return "Professional E";
	case PRODUCT_ENTERPRISE_E: return "Enterprise E";
	case PRODUCT_ULTIMATE_E: return "Ultimate E";
	case PRODUCT_UNLICENSED: return "Unlicensed";
	}

	return "(Unknown product type)";
}

//============================================================================

bool Cmd_GetOS(ScriptContext& context, void*)
{
	SYSTEM_INFO si;
	GetSystemInfo(&si);

	// Get OS version
	std::string strOS;
	OSVERSIONINFOEX osInfo;
	ZeroMemory(&osInfo, sizeof(osInfo));
	osInfo.dwOSVersionInfoSize = sizeof(osInfo);
	if(!GetVersionEx((OSVERSIONINFO*)&osInfo))
		strOS = "Unknown Operating System";
	else
	{
		// Win 9x
		if(osInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
		{
			if((osInfo.dwMajorVersion == 4) && (osInfo.dwMinorVersion == 0))
				strOS = "Windows 95";
			else if((osInfo.dwMajorVersion == 4) && (osInfo.dwMinorVersion == 10))
				strOS = "Windows 98";
			else if((osInfo.dwMajorVersion == 4) && (osInfo.dwMinorVersion == 90))
				strOS = "Windows ME";
			else
				strOS = "Unknown Win9x OS";
		}

		// Win NT
		else if(osInfo.dwPlatformId == VER_PLATFORM_WIN32_NT)
		{
			if((osInfo.dwMajorVersion == 4) && (osInfo.dwMinorVersion == 0))
				strOS = "Windows NT 4.0";
			else if(osInfo.dwMajorVersion == 6 && osInfo.dwMinorVersion == 0)
			{
				if(osInfo.wProductType == VER_NT_WORKSTATION)
					strOS = "Windows Vista ";
				else
					strOS = "Windows Server 2008 ";

				// Call GetProductInfo() to get extended version info
				LPGetProductInfo pfnGetProductInfo = (LPGetProductInfo)GetProcAddress(
					GetModuleHandle("kernel32.dll"), "GetProductInfo");
				if(pfnGetProductInfo)
				{
					DWORD dwType=0;
					pfnGetProductInfo(osInfo.dwMajorVersion, osInfo.dwMinorVersion, 0, 0, &dwType);
					strOS += GetProductType(dwType);
				}

				if(si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64)
					strOS += " 64-bit";
				else if(si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_INTEL)
					strOS += " 32-bit";
			}
			else if(osInfo.dwMajorVersion == 6 && osInfo.dwMinorVersion == 1)
			{
				if(osInfo.wProductType == VER_NT_WORKSTATION)
					strOS = "Windows 7 ";
				else
					strOS = "Windows Server 2008 R2 ";

				// Call GetProductInfo() to get extended version info
				LPGetProductInfo pfnGetProductInfo = (LPGetProductInfo)GetProcAddress(
					GetModuleHandle("kernel32.dll"), "GetProductInfo");
				if(pfnGetProductInfo)
				{
					DWORD dwType;
					pfnGetProductInfo(osInfo.dwMajorVersion, osInfo.dwMinorVersion, 0, 0, &dwType);
					strOS += GetProductType(dwType);
				}

				if(si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64)
					strOS += " 64-bit";
				else if(si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_INTEL)
					strOS += " 32-bit";
			}
			else if(osInfo.dwMajorVersion == 6 && osInfo.dwMinorVersion == 2)
			{
				if(osInfo.wProductType == VER_NT_WORKSTATION)
					strOS = "Windows 8 ";
				else
					strOS = "Windows Server 2012 ";

				// Call GetProductInfo() to get extended version info
				LPGetProductInfo pfnGetProductInfo = (LPGetProductInfo)GetProcAddress(
					GetModuleHandle("kernel32.dll"), "GetProductInfo");
				if(pfnGetProductInfo)
				{
					DWORD dwType;
					pfnGetProductInfo(osInfo.dwMajorVersion, osInfo.dwMinorVersion, 0, 0, &dwType);
					strOS += GetProductType(dwType);
				}

				if(si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64)
					strOS += " 64-bit";
				else if(si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_INTEL)
					strOS += " 32-bit";
			}
			else if(osInfo.dwMajorVersion == 6 && osInfo.dwMinorVersion == 3)
			{
				if(osInfo.wProductType == VER_NT_WORKSTATION)
					strOS = "Windows 8.1 ";
				else
					strOS = "Windows Server 2012 R2 ";

				// Call GetProductInfo() to get extended version info
				LPGetProductInfo pfnGetProductInfo = (LPGetProductInfo)GetProcAddress(
					GetModuleHandle("kernel32.dll"), "GetProductInfo");
				if(pfnGetProductInfo)
				{
					DWORD dwType;
					pfnGetProductInfo(osInfo.dwMajorVersion, osInfo.dwMinorVersion, 0, 0, &dwType);
					strOS += GetProductType(dwType);
				}

				if(si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64)
					strOS += " 64-bit";
				else if(si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_INTEL)
					strOS += " 32-bit";
			}
			else if(osInfo.dwMajorVersion == 5 && osInfo.dwMinorVersion == 2)
			{
				if(GetSystemMetrics(SM_SERVERR2))
					strOS = "Windows Server 2003 R2, ";
				else if(osInfo.wSuiteMask & VER_SUITE_STORAGE_SERVER)
					strOS = "Windows Storage Server 2003";
				else if(osInfo.wProductType & VER_NT_WORKSTATION &&
					si.wProcessorArchitecture==PROCESSOR_ARCHITECTURE_AMD64)
				{
					strOS = "Windows XP Professional x64 Edition";
				}
				else
					strOS = "Windows Server 2003, ";

				// Test for the server type.
				if(osInfo.wProductType != VER_NT_WORKSTATION)
				{
					if(si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_IA64)
					{
						if(osInfo.wSuiteMask & VER_SUITE_DATACENTER)
							strOS += "Datacenter Edition for Itanium-based Systems";
						else if(osInfo.wSuiteMask & VER_SUITE_ENTERPRISE )
							strOS += "Enterprise Edition for Itanium-based Systems";
					}
					else if(si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64)
					{
						if(osInfo.wSuiteMask & VER_SUITE_DATACENTER)
							strOS += "Datacenter x64 Edition";
						else if( osInfo.wSuiteMask & VER_SUITE_ENTERPRISE)
							strOS += "Enterprise x64 Edition";
						else
							strOS += "Standard x64 Edition";
					}
					else
					{
						if(osInfo.wSuiteMask & VER_SUITE_COMPUTE_SERVER)
							strOS += "Compute Cluster Edition";
						else if(osInfo.wSuiteMask & VER_SUITE_DATACENTER)
							strOS += "Datacenter Edition";
						else if(osInfo.wSuiteMask & VER_SUITE_ENTERPRISE)
							strOS += "Enterprise Edition";
						else if (osInfo.wSuiteMask & VER_SUITE_BLADE )
							strOS += "Web Edition";
						else
							strOS += "Standard Edition";
					}
				}
			}
			else if(osInfo.dwMajorVersion == 5 && osInfo.dwMinorVersion == 0)
			{
				strOS = "Windows 2000 ";
				if(osInfo.wProductType == VER_NT_WORKSTATION)
				{
					strOS += "Professional";
				}
				else
				{
					if(osInfo.wSuiteMask & VER_SUITE_DATACENTER)
						strOS += "Datacenter Server";
					else if(osInfo.wSuiteMask & VER_SUITE_ENTERPRISE)
						strOS += "Advanced Server";
					else
						strOS += "Server";
				}
			}
			else if(osInfo.dwMajorVersion == 5 && osInfo.dwMinorVersion == 1)
			{
				strOS = "Windows XP ";
				if(osInfo.wSuiteMask & VER_SUITE_PERSONAL)
					strOS += "Home Edition";
				else
					strOS += "Professional";
			}
			else
				strOS = "Unknown Windows OS";
		}

		// Unknown
		else
			strOS = "Unknown Operating System";

		strOS += ' ';
		strOS += osInfo.szCSDVersion;
	}

	context.AddRetval(strOS.c_str());
	return true;
}

//============================================================================
