//============================================================================
// GetTime.cpp - Standard library function: getTime()
//============================================================================
#include "Plugin/ScriptContext.h"
#include <stdio.h>
#include <time.h>
#include <windows.h>
#include <string>

//============================================================================

static const char* s_szWeekdaysShort[] = {
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

static const char* s_szWeekdaysLong[] = {
	"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"
};

static const char* s_szMonthsShort[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

static const char* s_szMonthsLong[] = {
	"January", "February", "March", "April", "May", "June", "July", "August",
	"September", "October", "November", "December"
};

//============================================================================

bool Cmd_GetTime(ScriptContext& context, void*)
{
	// Get format strinng
	if(context.GetNumArgs() != 1)
	{
		context.SetError("getTime() expects one parameter");
		return false;
	}
	ScriptVar fmt = context.GetArg(0);
	if(fmt.GetType() != ScriptVar::TYPE_String)
	{
		context.SetError("getTime() expects a string as the first parameter");
		return false;
	}
	std::string strFormat = fmt.ToString();

	// Get current time
	__time64_t currTime = _time64(NULL);
	const tm* pTime = _localtime64(&currTime);
	if(!pTime)
	{
		context.SetError("Failed to read current system time");
		return false;
	}

	// See if it's a leap year
	bool bLeapYear;
	{
		int nYear = pTime->tm_year+1900;
		if(nYear%400 == 0) bLeapYear = true;
		else if(nYear%100 == 0) bLeapYear = false;
		else if(nYear%4 == 0) bLeapYear = true;
		else bLeapYear = false;
	}

	// Construct string
	std::string strTime;
	char szBuff[256];
	int nBuff;
	for(size_t i=0; i<strFormat.length(); ++i)
	{
		char ch = strFormat[i];
		switch(ch)
		{
			//
			// Day:
			//
			case 'd':
				sprintf(szBuff, "%02d", pTime->tm_mday);
				strTime += szBuff;
				break;

			case 'D':
				strTime += s_szWeekdaysShort[pTime->tm_wday];
				break;

			case 'j':
				sprintf(szBuff, "%d", pTime->tm_mday);
				strTime += szBuff;
				break;

			case 'l':
				strTime += s_szWeekdaysLong[pTime->tm_wday];
				break;

			case 'S':
				if(pTime->tm_mday==1 || pTime->tm_mday==11 || pTime->tm_mday==21 || pTime->tm_mday==31)
					strTime += "st";
				else if(pTime->tm_mday==2 || pTime->tm_mday==22)
					strTime += "nd";
				else if(pTime->tm_mday==3 || pTime->tm_mday==23)
					strTime += "rd";
				else
					strTime += "th";
				break;

			case 'w':
				sprintf(szBuff, "%d", pTime->tm_wday);
				strTime += szBuff;
				break;

			case 'z':
				sprintf(szBuff, "%d", pTime->tm_yday);
				strTime += szBuff;
				break;

			//
			// Month:
			//
			case 'F':
				strTime += s_szMonthsLong[pTime->tm_mon];
				break;

			case 'm':
				sprintf(szBuff, "%02d", pTime->tm_mon+1);
				strTime += szBuff;
				break;

			case 'M':
				strTime += s_szMonthsShort[pTime->tm_mon];
				break;

			case 'n':
				sprintf(szBuff, "%d", pTime->tm_mon+1);
				strTime += szBuff;
				break;

			case 't':
				if(pTime->tm_mon==3 || pTime->tm_mon==5 || pTime->tm_mon==8 || pTime->tm_mon==10)
					nBuff = 30;
				else if(pTime->tm_mon==1)
					nBuff = bLeapYear ? 29 : 28;
				else
					nBuff = 31;
				sprintf(szBuff, "%d", nBuff);
				strTime += szBuff;
				break;

			//
			// Year:
			//
			case 'L':
				strTime += bLeapYear ? "1" : "0";
				break;

			case 'Y':
				sprintf(szBuff, "%02d", pTime->tm_year+1900);
				strTime += szBuff;
				break;

			case 'y':
				sprintf(szBuff, "%02d", pTime->tm_year%100);
				strTime += szBuff;
				break;

			//
			// Time:
			//
			case 'a':
				strTime += pTime->tm_hour < 12 ? "am" : "pm";
				break;

			case 'A':
				strTime += pTime->tm_hour < 12 ? "AM" : "PM";
				break;

			case 'g':
				if(pTime->tm_hour == 0)
					strTime += "12";
				else if(pTime->tm_hour < 13)
				{
					sprintf(szBuff, "%d", pTime->tm_hour);
					strTime += szBuff;
				}
				else
				{
					sprintf(szBuff, "%d", pTime->tm_hour-12);
					strTime += szBuff;
				}
				break;

			case 'G':
				sprintf(szBuff, "%d", pTime->tm_hour);
				strTime += szBuff;
				break;

			case 'h':
				if(pTime->tm_hour == 0)
					strTime += "12";
				else if(pTime->tm_hour < 13)
				{
					sprintf(szBuff, "%02d", pTime->tm_hour);
					strTime += szBuff;
				}
				else
				{
					sprintf(szBuff, "%02d", pTime->tm_hour-12);
					strTime += szBuff;
				}
				break;

			case 'H':
				sprintf(szBuff, "%02d", pTime->tm_hour);
				strTime += szBuff;
				break;

			case 'i':
				sprintf(szBuff, "%02d", pTime->tm_min);
				strTime += szBuff;
				break;

			case 's':
				sprintf(szBuff, "%02d", pTime->tm_sec);
				strTime += szBuff;
				break;

			case 'u': // Milli-seconds
					sprintf(szBuff, "%d", timeGetTime());
					strTime += szBuff;
				break;

			//
			// Timezone:
			//
			case 'I':
				strTime += pTime->tm_isdst ? "1" : "0";
				break;

			//
			// Full Date / Time:
			//
			case 'r':
				{
					// Remove newline
					std::string strTmp = asctime(pTime);
					strTmp.erase(strTmp.end()-1);
					strTime += strTmp;
				}
				break;

			case 'U':
				sprintf(szBuff, "%I64d", currTime);
				strTime += szBuff;
				break;

			//
			// Unknown:
			//
			default:
				strTime += ch;
				break;
		}
	}

	context.AddRetval(strTime.c_str());
	return true;
}

//============================================================================
