//============================================================================
// SetIcon.cpp - Standard library function: setIcon() / setWindowTitle()
//============================================================================
#include "Plugin/ScriptContext.h"
#include "App/App.h"
#include <windows.h>

//============================================================================

bool Cmd_SetIcon(ScriptContext& context, void*)
{
	// Get format strinng
	if(context.GetNumArgs() > 1)
	{
		context.SetError("setIcon() expects one parameter (Icon filename), or none");
		return false;
	}

	HANDLE hIcon = NULL;
	if(context.GetNumArgs() == 1)
	{
		ScriptVar filename = context.GetArg(0);
		if(filename.GetType() != ScriptVar::TYPE_String)
		{
			context.SetError("setIcon() expects a string as the first parameter (Icon filename)");
			return false;
		}
		std::string strFilename = filename.ToString();

		// Load the icon
		hIcon = LoadImage(NULL, strFilename.c_str(), IMAGE_ICON, 0, 0,
			LR_LOADFROMFILE | LR_DEFAULTSIZE);
		if(!hIcon)
		{
			char szBuff[32];
			sprintf(szBuff, "%d", GetLastError());
			std::string strError = "Failed to load icon '" + strFilename +
				"'. Error: " + std::string(szBuff);
			context.SetError(strError.c_str());
			return false;
		}
	}

	// Set it as the window icon
	CApp::GetInstance()->SetIcon((HICON)hIcon);
	return true;
}

//============================================================================

bool Cmd_SetWindowTitle(ScriptContext& context, void*)
{
	// Get format strinng
	if(context.GetNumArgs() > 1)
	{
		context.SetError("setWindowTitle() expects one parameter (New title), or zero");
		return false;
	}

	const char* szTitle = NULL;
	std::string strTitle;
	if(context.GetNumArgs() == 1)
	{
		ScriptVar filename = context.GetArg(0);
		if(filename.GetType() != ScriptVar::TYPE_String)
		{
			context.SetError("setWindowTitle() expects a string as the first parameter (Window title)");
			return false;
		}
		strTitle = filename.ToString();
		szTitle = strTitle.c_str();
	}

	// Set it as the window icon
	CApp::GetInstance()->SetTitle(szTitle);
	return true;
}

//============================================================================
