//=================================================================================================
// StandardLib.h - Standard library of functions
//=================================================================================================

#ifndef __STANDARDLIB_H__
#define __STANDARDLIB_H__

#include <windows.h>
#include <stdio.h>
#include <string>
#include <vector>

class CScriptManager;

//=================================================================================================

struct HTTPGet
{
	std::string strScriptType;			// Script type that registered this get
	std::string strAddress;				// E.g. http://www.foo.com
	std::string strHeader;				// HTTP header
	std::string strContent;				// Page content
	std::string strContentFilename;		// Page content filename (Or "")
	FILE* pContentFile;					// FILE* containing content (Or NULL)
	std::string strError;				// Error message, or ""
	std::string strCallback;			// Script function to call on completion

	// Updated by worker thread. Do not modify!
	volatile bool bHeaderComplete;		// Is the HTTP header complete?
	volatile int nResponseCode;			// HTTP response code
	volatile int nContentLength;		// From Content-Length field
	volatile size_t nBytesTransferred;	// Bytes of content transferred (Not including header)
	volatile int nTransferSpeed;		// Transfer speed in bytes/s

	HANDLE hThread;						// Handle to worker thread
	UINT_PTR nUserData;					// User data
};

struct StandardLibState
{
	std::vector<HTTPGet*> vGets;	// Pending HTTP GET requests
	HANDLE hThreadTerminate;		// Signalled when all threads should exit
};

//=================================================================================================

HTTPGet* StartHTTPGet(const std::string& strAddress, const std::string& strFilename,
	HANDLE hTerminate, std::string& strError);
bool RegisterStandardLibrary(CScriptManager& sm);
void TickStandardLibrary(CScriptManager& sm);
void CleanupStandardLibrary(CScriptManager& sm);

//=================================================================================================

#endif // __STANDARDLIB_H__
