//============================================================================
// TA2Stream.cpp - Stringstream-like class for composing TA2 tag strings
//============================================================================
#include "TA2Stream.h"
#include <windows.h>


//============================================================================
CTA2Stream::CTA2Stream()
{
}

CTA2Stream::~CTA2Stream()
{
}

//============================================================================
// Functions
std::string ParseTA2Tags(const std::string& str)
{
	std::string strRet(str);
	for(size_t nPos=0; nPos<strRet.length(); ++nPos)
	{
		// Skip to the next "<-"
		if(strRet.length() < nPos+4)	// Too short for the shortest tag
			break;
		if(strRet[nPos] != '<')
			continue;
		if(strRet[nPos+1] != '-')
			continue;

		// Does it look recognisible?
		char ch = strRet[nPos+2];
		if((ch == 'b') || (ch == 'B') || (ch == L'1'))
		{
			if(strRet[nPos+3] != '>')
				continue;
			strRet[nPos] = (char)(unsigned char)0xc0 | 0x20;
			strRet.erase(nPos+1, 3);
			continue;
		}
		if((ch == 'i') || (ch == 'I'))
		{
			if(strRet[nPos+3] != '>')
				continue;
			strRet[nPos] = (char)(unsigned char)0xc0 | 0x10;
			strRet.erase(nPos+1, 3);
			continue;
		}
		if((ch == 'u') || (ch == 'U'))
		{
			if(strRet[nPos+3] != '>')
				continue;
			strRet[nPos] = (char)(unsigned char)0xc0 | 0x08;
			strRet.erase(nPos+1, 3);
			continue;
		}
		if((ch == 's') || (ch == 'S'))
		{
			if(strRet[nPos+3] != '>')
				continue;
			strRet[nPos] = (char)(unsigned char)0xc0 | 0x02;
			strRet.erase(nPos+1, 3);
			continue;
		}
		if((ch == 'c') || (ch == 'C'))
		{
			if((strRet.length() < nPos+10) || (strRet[nPos+9] != '>'))
				continue;
			strRet[nPos] = (char)(unsigned char)0xc0 | 0x04;
			strRet.erase(nPos+1, 2);
			strRet.erase(nPos+7, 1);
			continue;
		}
		if(ch == '3')
		{
			if((strRet.length() < nPos+5) || (strRet[nPos+4] != L'>'))
				break;
			char chColour = strRet[nPos+3];
			const char* szColour;
			if(chColour == '1') szColour = "ff0000";
			else if(chColour == '2') szColour = "00ff00";
			else if(chColour == '3') szColour = "ffff00";
			else if(chColour == '4') szColour = "0000ff";
			else if(chColour == '5') szColour = "ff00ff";
			else if(chColour == '6') szColour = "00ffff";
			else if(chColour == '7') szColour = "ffffff";
			else continue;

			strRet[nPos] = (char)(unsigned char)0xc0 | 0x04;
			strRet.insert(strRet.begin()+nPos+5, 2, ' ');
			memcpy(&strRet[nPos+1], szColour, 6);
			continue;
		}
		if(ch == '0')
		{
			if(strRet[nPos+3] != '>')
				continue;
			strRet[nPos] = (char)(unsigned char)0xbe;
			strRet.erase(nPos+1, 3);
			continue;
		}
		if(ch == '/')
		{
			if(strRet.length() < nPos+5)
				break;

			char ch = strRet[nPos+3];
			if((ch == 'b') || (ch == 'B'))
			{
				if(strRet[nPos+4] != '>')
					continue;
				strRet[nPos] = (char)(unsigned char)0x80 | 0x20;
				strRet.erase(nPos+1, 4);
				continue;
			}
			if((ch == 'i') || (ch == 'I'))
			{
				if(strRet[nPos+4] != '>')
					continue;
				strRet[nPos] = (char)(unsigned char)0x80 | 0x10;
				strRet.erase(nPos+1, 4);
				continue;
			}
			if((ch == 'u') || (ch == 'U'))
			{
				if(strRet[nPos+4] != '>')
					continue;
				strRet[nPos] = (char)(unsigned char)0x80 | 0x08;
				strRet.erase(nPos+1, 4);
				continue;
			}
			if((ch == 's') || (ch == L'S'))
			{
				if(strRet[nPos+4] != '>')
					continue;
				strRet[nPos] = (char)(unsigned char)0x80 | 0x02;
				strRet.erase(nPos+1, 4);
				continue;
			}
		}
	}

	return strRet;
}
