//============================================================================
// Window.cpp - Window base class
//============================================================================
#include <windows.h>
#include <stdio.h>
#include "Window.h"
#include "Resource/Resource.h"

#define WINDOW_CLASSNAME "DruinkClientWindow"

// Fucking Microsoft headers
#ifdef _WIN64
#	define CAST_TYPE	(LONG_PTR)
#else
#	define CAST_TYPE	(LONG)(LONG_PTR)
#endif

//============================================================================
// Inlines
inline void ErrorBox(LPCSTR szError)
	{MessageBox(NULL, szError, "Error", MB_OK|MB_ICONERROR);}
inline void ErrorBox(HWND hWnd, LPCSTR szError)
	{MessageBox(hWnd, szError, "Error", MB_OK|MB_ICONERROR);}

//============================================================================
// Window class
CWindow::CWindow() :
	m_hInstance(NULL),
	m_hWnd(NULL),
	m_hAccelerator(NULL),
	m_strWindowTitle("Window"),
	m_bExit(false),
	m_bActive(false),
	m_nWidth(320),
	m_nHeight(200),
	m_nXPos(CW_USEDEFAULT),
	m_nYPos(CW_USEDEFAULT)
{
}

bool CWindow::Init(HINSTANCE hInstance)
{
TCHAR szBuff[256];

	// Register window class
	WNDCLASSEX wc;
	m_hInstance = hInstance;
	memset(&wc,0,sizeof(wc));
	wc.cbSize = sizeof(wc);
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = StaticWndProc;
	wc.hInstance = hInstance;
	wc.hIcon = GetIcon();
	wc.hCursor = GetCursor();
	wc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wc.lpszClassName = WINDOW_CLASSNAME;
	wc.hIconSm = wc.hIcon;
	if(!RegisterClassEx(&wc))
	{
		sprintf(szBuff, "RegisterClassEx() failed. GetLastError() returns %d", GetLastError());
		ErrorBox(szBuff);
		return false;
	}

	// Load accelerator
	m_hAccelerator = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCELERATOR));

	// Create window
	DWORD dwStyle = WS_OVERLAPPEDWINDOW;
	m_hWnd = CreateWindowEx(0, WINDOW_CLASSNAME, m_strWindowTitle.c_str(),
		dwStyle, m_nXPos, m_nYPos, m_nWidth, m_nHeight, NULL, NULL, hInstance, this);
	if(!m_hWnd)
	{
		sprintf(szBuff, "CreateWindowEx() failed. GetLastError() returns %d", GetLastError());
		UnregisterClass(WINDOW_CLASSNAME, hInstance);
		ErrorBox(szBuff);
		return false;
	}

	// Done
	return true;
}

void CWindow::Shutdown()
{
	if(m_hWnd)
	{
		DestroyWindow(m_hWnd);
		m_hWnd = NULL;
	}
	UnregisterClass(WINDOW_CLASSNAME, m_hInstance);
	m_strWindowTitle = "Window";
	m_hInstance = NULL;
	m_bExit = false;
	m_bActive = false;
	m_nWidth = 320;
	m_nHeight = 200;
	m_nXPos = CW_USEDEFAULT;
	m_nYPos = CW_USEDEFAULT;
}

void CWindow::Go()
{
	while(PumpWindowMessages())
	{
		// Update
		Tick();
	}
}

bool CWindow::PumpWindowMessages()
{
MSG msg;

	if(InHighPerfMode())
	{
		while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if(msg.message == WM_QUIT)
				m_bExit = true;

			if(!m_hAccelerator || !TranslateAccelerator(m_hWnd, m_hAccelerator, &msg))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
	}
	else
	{
		BOOL bRet = GetMessage(&msg, NULL, 0, 0);
		if(bRet > 0)
		{
			if(!m_hAccelerator || !TranslateAccelerator(m_hWnd, m_hAccelerator, &msg))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		else
			m_bExit = true;
	}

	return !m_bExit;
}

LRESULT CALLBACK CWindow::StaticWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
CWindow* pParent;

	// Get pointer to app
	if(uMsg == WM_NCCREATE)
	{
		pParent = (CWindow*)((LPCREATESTRUCT)lParam)->lpCreateParams;
		SetWindowLongPtr(hWnd, GWLP_USERDATA, CAST_TYPE pParent);
	}
	else
	{
		pParent = (CWindow*)(LONG_PTR)GetWindowLongPtr(hWnd, GWLP_USERDATA);
		if(!pParent) return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}

	pParent->m_hWnd = hWnd;
	return pParent->WindowProc(uMsg, wParam, lParam);
}

LRESULT CWindow::WindowProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	return WndProc(uMsg, wParam, lParam);
}

LRESULT CWindow::WndProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_DESTROY:
		m_bExit = true;
		break;

	case WM_ACTIVATE:
		m_bActive = (LOWORD(wParam) != WA_INACTIVE);
		break;

	case WM_NOTIFY:
		OnNotify((NMHDR*)lParam);
		break;
	}

	return DefWindowProc(m_hWnd, uMsg, wParam, lParam);
}
