//============================================================================
// ScriptContextLua.h - Script context for Lua
//============================================================================

#ifndef __SCRIPTCONTEXTLUA_H__
#define __SCRIPTCONTEXTLUA_H__

#include "ScriptContextInterface/ScriptContextInterface.h"
#include <string>
#include <vector>

struct lua_State;
struct lua_Debug;

class ScriptContextLua : public ScriptContextInterface
{
public:
	ScriptContextLua();
	ScriptContextLua(lua_State* pState);
	virtual ~ScriptContextLua();

	//========================================================================
	// ScriptContext:
	virtual size_t GetNumArgs() const;
	virtual ScriptVar GetArg(size_t nIdx) const;
	virtual void AddRetval(const ScriptVar& var);
	virtual void AddArg(const ScriptVar& var);
	virtual void SetError(const char* szError);

	//========================================================================
	// ScriptContextInterface:
	virtual void Destroy();
	virtual const char* GetScriptType();
	virtual ScriptLoadResult LoadScript(const char* szScriptName);
	virtual bool LoadScript(const void* pData, size_t nLen);
	virtual bool ReloadScripts();
	virtual bool Register(LPScriptFunction pfnFunc, void* pParam, const char* szName);
	virtual bool Unregister(const char* szFuncName);
	virtual bool Call(const char* szFunction, size_t nArgs, const ScriptVar* pArgs,
		ScriptVar** ppRetvals, size_t& rnRetvals);
	virtual const char* GetError() const;

	//========================================================================
	// ScriptContextLua:
	bool Init();

private:
	static int StaticFunctionStub(lua_State* pState);
	static void StaticHookProc(lua_State* _pState, lua_Debug* _pDebug);

	// Helper function: Lua var -> EScriptVar
	static ScriptVar GetLuaVar(lua_State* pState, int nIdx);

	// Helper function: EScriptVar -> Lua var
	static void PushVar(lua_State* pState, const ScriptVar& var);

	static const char* StaticChunkReader(lua_State*, void* pData, size_t* pSize);
	static const char* StaticDataReader(lua_State*, void* pData, size_t* pSize);

private:
	static const size_t ms_nMaxScriptNameLen = 256;

	struct ScriptFunction
	{
		LPScriptFunction pfnFunction;
		void* pParam;
		char szName[ms_nMaxScriptNameLen];
		ScriptContextLua* pThis;
	};

	lua_State* m_pState;
	int m_nRetvals;
	size_t m_nArgs;
	bool m_bOwnState;

	std::vector<std::string> m_vLoadedScripts;
	std::vector<ScriptFunction*> m_vScriptFunctions;
	std::string m_strError;

	std::vector<ScriptVar> m_vReturnValues;
};

#endif // __SCRIPTCONTEXTLUA_H__
