//============================================================================
// Main.cpp - Currsong plugin
//============================================================================
#include "Plugin/Plugin.h"
#include "Plugin/ScriptContext.h"
#include "Main.h"
#include "Winamp/Winamp.h"
#include "WMP/WMP.h"
#include <algorithm>

//============================================================================

bool Cmd_GetCurrsong(ScriptContext& context, void*)
{
	// Init default attributes
	Attributes attribs;
	attribs.nTrack = 0;
	attribs.nLength = 0;
	attribs.nPosition = 0;
	attribs.nBitrate = 0;
	attribs.nStamplerate = 0;
	attribs.nChannels = 0;

	// Get current song
	if(!GetCurrSong_Winamp(attribs) && !GetCurrSong_WMP(attribs))
	{
		// Failed to get current song info
		// Do nothing, so we use the default values above
	}

	// Process requested parameters
	size_t nArgs = context.GetNumArgs();
	for(size_t i=0; i<nArgs; ++i)
	{
		// Get param as string
		ScriptVar param = context.GetArg(i);
		if(param.GetType() != ScriptVar::TYPE_String)
		{
			context.SetError("getCurrSong expects all string parameters");
			return false;
		}

		std::string strParam = param.ToString();
		std::transform(strParam.begin(), strParam.end(), strParam.begin(), tolower);
		if(strParam == "number")
			context.AddRetval(attribs.nTrack);
		else if(strParam == "artist")
			context.AddRetval(attribs.strArtist.c_str());
		else if(strParam == "title")
			context.AddRetval(attribs.strTitle.c_str());
		else if(strParam == "player")
			context.AddRetval(attribs.strPlayer.c_str());
		else if(strParam == "length")
			context.AddRetval(attribs.nLength);
		else if(strParam == "position")
			context.AddRetval(attribs.nPosition);
		else if(strParam == "status")
			context.AddRetval(attribs.strStatus.c_str());
		else if(strParam == "bitrate")
			context.AddRetval(attribs.nBitrate);
		else if(strParam == "samplerate")
			context.AddRetval(attribs.nStamplerate);
		else if(strParam == "channels")
			context.AddRetval(attribs.nChannels);
		else
		{
			std::string strError = "Unknown parameter passed to getCurrSong: \"" + strParam + "\"";
			context.SetError(strError.c_str());
			return false;
		}
	}

	return true;
}

//============================================================================

PLUGIN_LINKAGE size_t GetNumExportedFunctions(int nAPIVersion)
{
	if(nAPIVersion != PLUGINAPI_VERSION)
		return 0;
	return 1;
}

PLUGIN_LINKAGE void GetExportedFunction(size_t nIndex, ScriptFunction& funcInfo)
{
	if(nIndex == 0)
	{
		funcInfo.pfnFunction = Cmd_GetCurrsong;
		funcInfo.pParam = 0;
		strcpy(funcInfo.szName, "getCurrSong");
	}
}

//============================================================================
