//============================================================================
// Winamp.cpp - Winamp current song support
//============================================================================
#include "../Main.h"
#include "Winamp SDK.h"
#include "Winamp STP SDK.h"
#include <windows.h>
#include <sstream>

//============================================================================
// Locate music player window and extract song name
//============================================================================
void GetInfo(HWND hWnd, std::string& strNumber, std::string& strArtist, std::string& strTitle)
{
char* pBuff;
char szBuff[1024];
size_t nPos;

	// Get current text
	GetWindowText(hWnd, szBuff, 1024);
	pBuff = strstr(szBuff, " - Winamp");
	if(pBuff == NULL)
		return;

	// Trim off " - Winamp" text
	*pBuff = '\0';

	// Fill in track num, artist & title
	strNumber = szBuff;
	strArtist = szBuff;
	strTitle = szBuff;

	// Get track number
	nPos = strNumber.find('.');
	if(nPos == std::string::npos)
		strNumber = "";
	else
		strNumber.erase(strNumber.begin()+nPos, strNumber.end());

	// Remove track num from atrist & title
	if(nPos != std::string::npos)
	{
		if(strArtist.length() > (DWORD)nPos+1)
		{
			strArtist.erase(strArtist.begin(), strArtist.begin()+nPos+2);
			strTitle.erase(strTitle.begin(), strTitle.begin()+nPos+2);
		}
		else
		{
			strArtist.erase(strArtist.begin(), strArtist.begin()+nPos+1);
			strTitle.erase(strTitle.begin(), strTitle.begin()+nPos+1);
		}
	}

	// Split on " - "
	nPos = strArtist.find(" - ");
	if(nPos == std::string::npos)
	{
		strTitle = strArtist;
		strArtist = "";
	}
	else
	{
		strArtist.erase(strArtist.begin()+nPos, strArtist.end());
		strTitle.erase(strTitle.begin(), strTitle.begin()+nPos+3);
	}
}

bool GetCurrSong_Winamp(Attributes& rData)
{
HWND hWnd;
LRESULT nBuff;

	// Find winamp window
	hWnd = FindWindow("Winamp v1.x", NULL);
	if(!hWnd)
		return false;

	// Get playing status
	switch(SendMessage(hWnd, WM_USER, 0, WINAMP_ISPLAYING))
	{
	case 1: rData.strStatus = "Playing"; break;
	case 3: rData.strStatus = "Paused"; break;
	default: rData.strStatus = "Stopped"; break;
	}

	// Get track number, artist and title
	std::string strNumber;
	GetInfo(hWnd, strNumber, rData.strArtist, rData.strTitle);
	std::istringstream is(strNumber);
	if(!(is >> rData.nTrack))
		rData.nTrack = 0;

	// Get position
	nBuff = SendMessage(hWnd, WM_USER, 0, WINAMP_GETOUTPUTTIME);
	if(nBuff != -1)
		rData.nPosition = (int)(nBuff/1000);

	// Get length
	nBuff = SendMessage(hWnd, WM_USER, 1, WINAMP_GETOUTPUTTIME);
	if(nBuff != -1)
		rData.nLength = (int)nBuff;

	// Get bitrate
	nBuff = SendMessage(hWnd, WM_USER, 1, WINAMP_GETINFO);
	if(nBuff != -1)
		rData.nBitrate = (int)nBuff;

	// Get samplerate
	nBuff = SendMessage(hWnd, WM_USER, 0, WINAMP_GETINFO);
	if(nBuff != -1)
		rData.nStamplerate = (int)nBuff;

	// Get channels
	nBuff = SendMessage(hWnd, WM_USER, 2, WINAMP_GETINFO);
	if(nBuff != -1)
		rData.nChannels = (int)nBuff;

	rData.strPlayer = "Winamp";
	return true;
}
