-- Weather.lua
-- Script for /weather command

-- Weather formatting and displaying. temp is a number
function displayWeather(location, temp, conditions, humidity)

	-- Ok, we have the temperature and weather, spew it out:
	local weatherString = ".me looks out the window at " .. location .. " and feels it's <-c00ff00><-b>" .. temp .. "'C<-0><-cffff00>, <-0><-b>" .. conditions .. ", <-0><-c00ff00><-b>" .. humidity .. " <-0><-b>humidity";
	sendln(weatherString);
end


-- httpGet callback
function weatherCallback(addr, response, header, content, error)

	-- Check for errors
	if response ~= 200 then
		println("<-cff0000>Failed to get current weather. Error: " .. error);
		return;
	end

	-- Convert content to lower case
	local lowerContent = string.lower(content);
	
	-- Loop through the <title> tags till we find one with the weather string in it
	local location, temp, conditions, humidity;
	local startPos = 1;
	local contentLength = string.len(lowerContent);
	repeat
		-- Find <title> tag start and end pos
		local tagStart = string.find(lowerContent, "<title>", startPos);
		if tagStart== nil then
			println("<-cff0000>Failed to get current weather. Could not find weather in returned data.");
			return;
		end
		local tagEnd = string.find(lowerContent, "</title>", tagStart+7);
		if tagEnd == nil then
			println("<-cff0000>Failed to get current weather. Could not find weather in returned data.");
			return;
		end
		startPos = tagEnd+8;

		-- Is this the tag we're looking for? (Contains some data after the word Weather
		local tag = string.sub(lowerContent, tagStart+7, tagEnd-1);
		local weatherPos = string.find(tag, "weather");
		if weatherPos ~= nil and string.len(tag)-(weatherPos+7) > 2 then
			-- Looks reasonable, grab location (Everything before first comma)
			local commaPos = string.find(tag, ",");
			if commaPos ~= nil then
				location = string.sub(content, tagStart+7, tagStart+7+commaPos-1);
			end

			-- Trim off everything up to and including "weather :: "
			local weatherString = string.sub(content, tagStart+7+weatherPos+10, tagEnd-1);
			local weatherStringLwr = string.lower(weatherString);
			
			-- Extract temp
			local tempPos = string.find(weatherStringLwr, "c");
			if tempPos ~= nil then
				temp = tonumber(string.sub(weatherString, 1, tempPos-1));
				conditions = string.sub(weatherString, tempPos+2);
			end
		end
	until temp ~= nill;
	
	-- Attempt to find the humidity in this piece of crap
	local humidStart = string.find(lowerContent, ">", string.find(lowerContent, "id=\"humidity\""));
	if humidStart ~= nil then
		local humidEnd = string.find(lowerContent, "<", humidStart);
		humidity = string.sub(lowerContent, humidStart + 1, humidEnd - 1);
	end

	-- Ok, we have the temperature and weather, spew it out
	displayWeather(location, temp, conditions, humidity);
end

-- Command to register
function cmdWeather(params)
	params = string.gsub(params, "^%s*(.-)%s*$", "%1");
	
	-- Grab page
	local configValue = "WeatherURL" .. params;
	local addr = getConfigValue(configValue);
	httpGet(addr, "weatherCallback");
end

register_command("weather", "cmdWeather");